#!/usr/bin/env python
"""
_Step.Executor.DQMUpload_

Implementation of an Executor for a DQMUpload step

"""
from __future__ import print_function

from future.utils import viewitems

import logging
import os
import sys
from io import BytesIO
from functools import reduce
from gzip import GzipFile
from hashlib import md5
from mimetypes import guess_type

try:
    # python2
    import urllib2
    HTTPError = urllib2.HTTPError
    OpenerDirector = urllib2.OpenerDirector
    Request = urllib2.Request
    ProxyHandler = urllib2.ProxyHandler
except:
    # python3
    import urllib.request
    HTTPError = urllib.error.HTTPError
    OpenerDirector = urllib.request.OpenerDirector
    Request = urllib.request.Request
    ProxyHandler = urllib.request.ProxyHandler

from Utils.Utilities import decodeBytesToUnicode, encodeUnicodeToBytesConditional, encodeUnicodeToBytes
from Utils.PythonVersion import PY3, PY2

from WMCore.FwkJobReport.Report import Report
from WMCore.Services.HTTPS.HTTPSAuthHandler import HTTPSAuthHandler
from WMCore.WMSpec.Steps.Executor import Executor
from WMCore.WMSpec.Steps.WMExecutionFailure import WMExecutionFailure


class DQMUpload(Executor):
    """
    _DQMUpload_

    Execute a DQMUpload Step

    """

    def pre(self, emulator=None):
        """
        _pre_

        Pre execution checks

        """
        # Are we using an emulator?
        if emulator is not None:
            return emulator.emulatePre(self.step)

        logging.info("Steps.Executors.%s.pre called", self.__class__.__name__)
        return None

    def execute(self, emulator=None):
        """
        _execute_

        """
        # Are we using emulators again?
        if emulator is not None:
            return emulator.emulate(self.step, self.job)

        logging.info("Steps.Executors.%s.execute called", self.__class__.__name__)

        if self.step.upload.proxy:
            try:
                self.stepSpace.getFromSandbox(self.step.upload.proxy)
            except Exception as ex:
                # Let it go, it wasn't in the sandbox. Then it must be
                # somewhere else
                del ex

        # Search through steps for analysis files
        for step in self.stepSpace.taskSpace.stepSpaces():
            if step == self.stepName:
                # Don't try to parse your own report; it's not there yet
                continue
            stepLocation = os.path.join(self.stepSpace.taskSpace.location, step)
            logging.info("Beginning report processing for step %s", step)
            reportLocation = os.path.join(stepLocation, 'Report.pkl')
            if not os.path.isfile(reportLocation):
                logging.error("Cannot find report for step %s in space %s", step, stepLocation)
                continue

            # First, get everything from a file and 'unpersist' it
            stepReport = Report()
            stepReport.unpersist(reportLocation, step)

            # Don't upload nor stage out files from bad steps.
            if not stepReport.stepSuccessful(step):
                continue

            # Pulling out the analysis files from each step
            analysisFiles = stepReport.getAnalysisFilesFromStep(step)

            # Working on analysis files
            for analysisFile in analysisFiles:
                # only deal with DQM files
                if analysisFile.FileClass == "DQM":
                    # uploading file to the server
                    self.httpPost(os.path.join(stepLocation,
                                               os.path.basename(analysisFile.fileName)))

            # Am DONE with report
            # Persist it
            stepReport.persist(reportLocation)

        return

    def post(self, emulator=None):
        """
        _post_

        Post execution checkpointing

        """
        # Another emulator check
        if emulator is not None:
            return emulator.emulatePost(self.step)

        logging.info("Steps.Executors.%s.post called", self.__class__.__name__)
        return None

    #
    # for the latest DQM upload code see https://github.com/rovere/dqmgui/blob/master/bin/visDQMUpload
    #

    def httpPost(self, filename):
        """
        _httpPost_

        perform an HTTP POST operation to a webserver

        """
        args = {}

        # Preparing a checksum
        blockSize = 0x10000

        def upd(m, data):
            m.update(data)
            return m

        with open(filename, 'rb') as fd:
            contents = iter(lambda: fd.read(blockSize), b'')
            m = reduce(upd, contents, md5())

        args['checksum'] = 'md5:%s' % m.hexdigest()
        # args['checksum'] = 'md5:%s' % md5.new(filename).read()).hexdigest()
        args['size'] = os.path.getsize(filename)

        msg = "HTTP Upload is about to start:\n"
        msg += " => URL: %s\n" % self.step.upload.URL
        msg += " => Filename: %s\n" % filename
        logging.info(msg)

        try:
            logException = True
            for uploadURL in self.step.upload.URL.split(';'):
                (headers, data) = self.upload(uploadURL, args, filename)
                msg = '  Status code: %s\n' % headers.get("Dqm-Status-Code", None)
                msg += '  Message: %s\n' % headers.get("Dqm-Status-Message", None)
                msg += '  Detail: %s\n' % headers.get("Dqm-Status-Detail", None)
                msg += '  Data: %s\n' % str(data)
                if int(headers.get("Dqm-Status-Code", 400)) >= 400:
                    logException = False
                    raise Exception(msg)
                else:
                    msg = 'HTTP upload finished succesfully with response:\n' + msg
                    logging.info(msg)
        except HTTPError as ex:
            msg = 'HTTP upload failed with response:\n'
            msg += '  Status code: %s\n' % ex.hdrs.get("Dqm-Status-Code", None)
            msg += '  Message: %s\n' % ex.hdrs.get("Dqm-Status-Message", None)
            msg += '  Detail: %s\n' % ex.hdrs.get("Dqm-Status-Detail", None)
            msg += '  Error: %s\n' % str(ex)
            logging.exception(msg)
            raise WMExecutionFailure(70318, "DQMUploadFailure", msg)
        except Exception as ex:
            msg = 'HTTP upload failed! Error:\n%s' % str(ex)
            if logException:
                logging.exception(msg)
            else:
                logging.error(msg)
            raise WMExecutionFailure(70318, "DQMUploadFailure", msg)

        return

    def filetype(self, filename):
        return guess_type(filename)[0] or 'application/octet-stream'

    def encode(self, args, files):
        """
        Encode form (name, value) and (name, filename, type) elements into
        multi-part/form-data. We don't actually need to know what we are
        uploading here, so just claim it's all text/plain.
        """
        boundary = b'----------=_DQM_FILE_BOUNDARY_=-----------'
        (body, crlf) = (b'', b'\r\n')
        for (key, value) in viewitems(args):
            logging.debug("encode value - %s, %s", type(value), value)
            if PY2:
                payload = str(value)
            elif PY3:
                payload = value
                if not isinstance(payload, bytes):
                    payload = str(payload)
                payload = encodeUnicodeToBytes(payload)
                key = encodeUnicodeToBytes(key)
            logging.debug("encode payload - %s, %s", type(payload), payload)
            body += b'--' + boundary + crlf
            body += (b'Content-Disposition: form-data; name="%s"' % key) + crlf
            body += crlf + payload + crlf
        for (key, filename) in viewitems(files):
            body += b'--' + boundary + crlf
            key = encodeUnicodeToBytesConditional(key, condition=PY3)
            filepath = encodeUnicodeToBytesConditional(os.path.basename(filename), condition=PY3)
            body += (b'Content-Disposition: form-data; name="%s"; filename="%s"'
                     % (key, filepath)) + crlf
            body += (b'Content-Type: %s' % encodeUnicodeToBytes(self.filetype(filename))) + crlf
            body += (b'Content-Length: %d' % os.path.getsize(filename)) + crlf
            logging.debug("encode body (without binary file) -%s, %s", type(body), body)
            with open(filename, "rb") as fd:
                body += crlf + fd.read() + crlf
            body += b'--' + boundary + b'--' + crlf + crlf
        return (b'multipart/form-data; boundary=' + boundary, body)

    def marshall(self, args, files, request):
        """
        Marshalls the arguments to the CGI script as multi-part/form-data,
        not the default application/x-www-form-url-encoded. This improves
        the transfer of the large inputs and eases command line invocation
        of the CGI script.
        """
        (contentType, body) = self.encode(args, files)
        request.add_header(b'Content-Type', contentType)
        if PY2:
            request.add_header('Content-Length', str(len(body)))
            request.add_data(body)
        elif PY3:
            request.data = body
        return

    def upload(self, url, args, filename):
        """
        _upload_

        Perform a file upload to the dqm server using HTTPS auth with the
        service proxy provided
        """
        ident = "WMAgent python/%d.%d.%d" % sys.version_info[:3]
        uploadProxy = self.step.upload.proxy or os.environ.get('X509_USER_PROXY', None)
        logging.info("Using proxy file: %s", uploadProxy)
        logging.info("Using CA certificate path: %s", os.environ.get('X509_CERT_DIR'))

        msg = "HTTP POST upload arguments:\n"
        for arg in args:
            msg += "  ==> %s: %s\n" % (arg, args[arg])
        logging.info(msg)

        handler = HTTPSAuthHandler(key=uploadProxy, cert=uploadProxy)
        opener = OpenerDirector()
        opener.add_handler(handler)

        # setup the request object
        url = decodeBytesToUnicode(url) if PY3 else encodeUnicodeToBytes(url)
        datareq = Request(url + '/data/put')
        datareq.add_header('Accept-encoding', 'gzip')
        datareq.add_header('User-agent', ident)
        self.marshall(args, {'file': filename}, datareq)

        if 'https://' in url:
            result = opener.open(datareq)
        else:
            opener.add_handler(ProxyHandler({}))
            result = opener.open(datareq)

        data = result.read()
        if result.headers.get('Content-encoding', '') == 'gzip':
            data = GzipFile(fileobj=BytesIO(data)).read()

        return (result.headers, data)



#!/usr/bin/env python
"""
_SetPhEDExStatus_

MySQL implementation of DBSBufferFiles.SetPhEDExStatus
"""

from WMCore.Database.DBFormatter import DBFormatter

class SetPhEDExStatus(DBFormatter):
    sql = "UPDATE dbsbuffer_file SET in_phedex = :status WHERE lfn = :lfn"

    def execute(self, lfns, status, conn = None, transaction = False):

        if not isinstance(lfns, list):
            lfns = [lfns]

        if len(lfns) < 1:
            return

        bindVars = []
        for lfn in lfns:
            bindVars.append({"lfn": lfn, "status": status})

        self.dbi.processData(self.sql, bindVars, conn = conn,
                             transaction = transaction)

        return



#!/usr/bin/env python
"""
_SetPhEDExStatus_

Oracle implementation of DBSBufferFiles.SetPhEDExStatus
"""

from WMComponent.DBS3Buffer.MySQL.DBSBufferFiles.SetPhEDExStatus \
      import SetPhEDExStatus as MySQLSetPhEDExStatus

class SetPhEDExStatus(MySQLSetPhEDExStatus):
    """
    _SetPhEDExStatus_

    Oracle implementation of SetPhEDExStatus
    """
    pass



#!/usr/bin/env python
"""
_StageOutImplV2_

Interface for Stage Out Plugins. All stage out implementations should
inherit this object and implement the methods accordingly

"""

from builtins import object

from WMCore.Storage.Execute import runCommandWithOutput as runCommand
from WMCore.Storage.StageOutError import StageOutError
import logging
import os

class StageOutImplV2(object):
    """
    _StageOutImplV2_

    """

    def __init__(self):
        self.numRetries = 5
        pass

    def doTransfer(self, fromPfn, toPfn, stageOut, pnn, command, options, protocol, checksum):
        """
            performs a transfer. stageOut tells you which way to go. returns the new pfn or
            raises on failure. StageOutError (and inherited exceptions) are for expected errors
            such as temporary connection failures. Anything else will be handled as an unexpected
            error and skip retrying with this plugin

            if stageOut is true:
                The fromPfn is the LOCAL FILE NAME on the node, without file://
                the toPfn is the target PFN, mapped from the LFN using the TFC or overrrides
            if stageOut is false:
                The toPfn is the LOCAL FILE NAME on the node, without file://
                the fromPfn is the source PFN, mapped from the LFN using the TFC or overrrides

            this behavior is because most transfer commands will switch their direction
            simply by swapping the order of the arguments. the stageOut flag is provided
            however, because sometimes you want to pass different command line args

        """
        raise NotImplementedError

    def doDelete(self, pfn, pnn, command, options, protocol  ):
        """
            deletes a file, raises on error
            StageOutError (and inherited exceptions) are for expected errors
            such as temporary connection failures. Anything else will be handled as an unexpected
            error and skip retrying with this plugin
        """
        raise NotImplementedError

    def prependFileProtocol(self, pfn):
        """
        _createSourceName_

        SRM uses file:/// urls

        """
        if os.path.isfile(pfn):
            return "file:%s" % pfn
        else:
            return pfn

    def runCommandFailOnNonZero(self, command):
        logging.info("Executing %s", command)
        (exitCode, output) = runCommand(command)
        if exitCode:
            logging.error("Error in file transfer:")
            logging.error("  Command executed was: %s", command )
            logging.error("  Output was: %s", output )
            logging.error("  Exit code was: %s", exitCode )
            raise StageOutError("Transfer failure")
        return (exitCode, output)

    def runCommandWarnOnNonZero(self, command):
        logging.info("Executing %s", command)
        (exitCode, output) = runCommand(command)
        if exitCode:
            logging.error("Error in file transfer..ignoring:")
            logging.error(output)
        return (exitCode, output)

    def generateCommandFromPreAndPostParts(self, pre,post,options):
        """
        simply put, this will return an array with [pre, options.split(), post]
        but if options is empty, won't put it in'
        """
        temp = pre
        if options != None:
            temp.extend(options.split())
        temp.extend(post)
        return temp



#!/usr/bin/env python
"""
_FixedDelay_

If the current time is beyond trigger_time, add
any unacquired jobs to
"""




from WMCore.JobSplitting.JobFactory import JobFactory
from WMCore.Services.UUIDLib import makeUUID
import time

class FixedDelay(JobFactory):
    """
    if the current time is past trigger_time,
    pull all the available files into a
    new job
    """
    def algorithm(self, *args, **kwargs):
        """
        _algorithm_

        A time delay job splitting algorithm, will shove all unacquired
        files into a new job if the trigger_time has been passed
        """

        #  //
        # // get the fileset
        #//
        fileset = self.subscription.getFileset()
        trigger_time = int(kwargs['trigger_time'])
        if (trigger_time < time.time()):
            availFiles = self.subscription.availableFiles()
            if (len(availFiles) == 0):
                # no files to acquire
                return []

            baseName = makeUUID()
            self.newGroup()
            self.newJob(name = '%s-endofrun' % (baseName,))

            for f in availFiles:
                self.currentJob.addFile(f)



#!/usr/bin/env python
"""
WorkQueuemanager component

Runs periodic tasks for WorkQueue
"""
from __future__ import print_function

import threading

from WMComponent.WorkQueueManager.WorkQueueManagerCleaner import WorkQueueManagerCleaner
from WMComponent.WorkQueueManager.WorkQueueManagerLocationPoller import WorkQueueManagerLocationPoller
from WMComponent.WorkQueueManager.WorkQueueManagerReqMgrPoller import WorkQueueManagerReqMgrPoller
from WMComponent.WorkQueueManager.WorkQueueManagerWMBSFileFeeder import WorkQueueManagerWMBSFileFeeder
from WMComponent.WorkQueueManager.WorkQueueManagerWorkPoller import WorkQueueManagerWorkPoller
from WMCore.Agent.Harness import Harness
from WMCore.WorkQueue.WorkQueueUtils import queueFromConfig, queueConfigFromConfigObject


class WorkQueueManager(Harness):
    """WorkQueuemanager component

    Runs periodic tasks for WorkQueue
    """

    def __init__(self, config):
        # call the base class
        Harness.__init__(self, config)
        self.config = queueConfigFromConfigObject(config)

    def preInitialization(self):
        print("WorkQueueManager.preInitialization")

        # Add event loop to worker manager
        myThread = threading.currentThread()
        pollInterval = self.config.WorkQueueManager.pollInterval
        dataLocationInterval = getattr(self.config.WorkQueueManager, "dataLocationInterval", 1 * 60 * 60)

        ### Global queue special functions
        if self.config.WorkQueueManager.level == 'GlobalQueue':

            # Get work from ReqMgr, report back & delete finished requests
            myThread.workerThreadManager.addWorker(
                    WorkQueueManagerReqMgrPoller(
                            queueFromConfig(self.config),
                            getattr(self.config.WorkQueueManager,
                                    'reqMgrConfig', {})
                    ),
                    pollInterval)

        ### local queue special function
        elif self.config.WorkQueueManager.level == 'LocalQueue':

            # pull work from parent queue
            myThread.workerThreadManager.addWorker(
                    WorkQueueManagerWorkPoller(queueFromConfig(self.config),
                                               self.config),
                    pollInterval)

            # inject acquired work into wmbs
            myThread.workerThreadManager.addWorker(
                    WorkQueueManagerWMBSFileFeeder(queueFromConfig(self.config),
                                                   self.config),
                    pollInterval)

        ### general functions

        # Data location updates
        myThread.workerThreadManager.addWorker(
                WorkQueueManagerLocationPoller(queueFromConfig(self.config),
                                               self.config),
                dataLocationInterval)

        # Clean finished work & apply end policies
        myThread.workerThreadManager.addWorker(
                WorkQueueManagerCleaner(queueFromConfig(self.config),
                                        self.config),
                pollInterval)

        return



#!/usr/bin/env python
"""
_GetFilesForMerge_

Oracle implementation of Subscription.GetFilesForMerge
"""

from WMCore.WMBS.MySQL.Subscriptions.GetFilesForMerge import GetFilesForMerge as GetFilesForMergeMySQL

class GetFilesForMerge(GetFilesForMergeMySQL):
    pass



#!/usr/bin/env python
"""
_WorkQueue_

WorkQueue provides functionality to queue large chunks of work,
thus acting as a buffer for the next steps in job processing

WMSpec objects are fed into the queue, split into coarse grained work units
and released when a suitable resource is found to execute them.

https://twiki.cern.ch/twiki/bin/view/CMS/WMCoreJobPool
"""
__all__ = []



#!/usr/bin/env python
# pylint: disable=E1101
# E1101: Allow imports from currentThread
"""
_ExecuteMaster_

Overseer object that traverses a task and invokes the type based executor
for each step

"""

from builtins import object

import logging
import os
import threading
import traceback

import WMCore.WMSpec.Steps.StepFactory as StepFactory
from WMCore.WMException import WMException
from WMCore.WMSpec.Steps.WMExecutionFailure import WMExecutionFailure
from WMCore.WMSpec.WMStep import WMStepHelper


class ExecuteMaster(object):
    """
    _ExecuteMaster_

    Traverse the given task and invoke the execute framework
    If an emulator is provided, then invoke the appropriate emulator
    instead of the executor

    """

    def __init__(self):
        pass

    def __call__(self, task, wmbsJob):
        """
        _operator(task)_

        Load and run executors for all steps in Task, if an emulator is
        available for that step, use it instead.

        """

        myThread = threading.currentThread

        try:
            myThread.watchdogMonitor.setupMonitors(task, wmbsJob)
            myThread.watchdogMonitor.notifyJobStart(task)
        except WMException:
            self.toTaskDirectory()
            raise
        except Exception as ex:
            msg = "Encountered unhandled exception while starting monitors:\n"
            msg += str(ex) + '\n'
            msg += str(traceback.format_exc()) + '\n'
            logging.error(msg)
            self.toTaskDirectory()
            raise WMExecutionFailure(msg)

        failureUpstream = False
        for step in task.steps().nodeIterator():
            try:
                helper = WMStepHelper(step)
                stepType = helper.stepType()
                stepName = helper.name()
                if failureUpstream:
                    # for chained steps, don't execute further steps if a
                    # failure has already happened
                    helper.addOverride("previousCmsRunFailure", True)

                executor = StepFactory.getStepExecutor(stepType)
                result = self.doExecution(executor, step, wmbsJob)
                logging.info("StepName: %s, StepType: %s, with result: %r", stepName, stepType, result)
                if result:  # can be either None, or the step exit code
                    failureUpstream = True
            except WMException as ex:
                msg = "Encountered error while running ExecuteMaster:\n"
                msg += str(ex) + "\n"
                logging.error(msg)
                self.toTaskDirectory()
                break
            except Exception as ex:
                msg = "Encountered error while running ExecuteMaster:\n"
                msg += str(ex) + "\n"
                msg += str(traceback.format_exc()) + "\n"
                self.toTaskDirectory()
                logging.error(msg)
                break

        try:
            myThread.watchdogMonitor.notifyJobEnd(task)
        except WMException:
            self.toTaskDirectory()
        except Exception as ex:
            msg = "Encountered unhandled exception while ending the job:\n"
            msg += str(ex) + '\n'
            msg += str(traceback.format_exc()) + '\n'
            logging.error(msg)
            self.toTaskDirectory()

        return

    def doExecution(self, executor, step, job):
        """
        _doExecution_

        Invoke the Executor for the step provided

        TODO: Add Monitoring thread & setup
        TODO: Exception Handling
        TODO: pre/post outcome can change the next execution task, need to
              ensure that this happens


        """
        myThread = threading.currentThread
        # Tell the watchdog that we're starting the step
        myThread.watchdogMonitor.notifyStepStart(step)

        self.toStepDirectory(step)
        executor.initialise(step, job)
        executionObject = executor
        error = False
        if executor.emulationMode:
            executionObject = executor.emulator

        preOutcome = executionObject.pre()
        if preOutcome is not None:
            logging.info("Pre Executor Task Change: %s", preOutcome)
            executor.saveReport()
            self.toTaskDirectory()
            myThread.watchdogMonitor.notifyStepEnd(step=step,
                                                   stepReport=executor.report)
            executor.saveReport()
            return preOutcome
        try:
            executor.report.setStepStartTime(stepName=executor.stepName)
            executionObject.execute()
        except WMExecutionFailure as ex:
            executor.diagnostic(ex.code, executor, ExceptionInstance=ex)
            executor.report.addError(executor.stepName, ex.code, "WMAgentStepExecutionError", str(ex))
            error = True
        except Exception as ex:
            logging.error("Exception occured when executing step")
            logging.error("Exception is %s", ex)
            logging.error("Traceback: ")
            logging.error(traceback.format_exc())
            executor.diagnostic(99109, executor, ExceptionInstance=ex)
            executor.report.addError(executor.stepName, 99109, "WMAgentStepExecutionError", str(ex))
            error = True
        executor.report.setStepStopTime(stepName=executor.stepName)
        # TODO: Handle generic Exception that indicates development/code errors
        executor.saveReport()

        postOutcome = executionObject.post()
        if postOutcome is not None:
            logging.info("Post Executor Task Change: %s", postOutcome)
            executor.saveReport()
            self.toTaskDirectory()
            myThread.watchdogMonitor.notifyStepEnd(step=step,
                                                   stepReport=executor.report)
            executor.saveReport()
            return postOutcome

        self.toTaskDirectory()

        # Okay, we're done, set the job to successful
        stepExitCode = executor.report.getExitCode()  # 0 is successful
        if not error and not stepExitCode:
            executor.report.setStepStatus(stepName=executor.stepName,
                                          status=0)
        executor.saveReport()

        # Tell the watchdog that we're done with the step
        myThread.watchdogMonitor.notifyStepEnd(step=step,
                                               stepReport=executor.report)
        executor.saveReport()

        return stepExitCode

    def toStepDirectory(self, step):
        """
        _toStepDirectory_

        Switch current working directory to the step location
        within WMTaskSpace

        """
        stepName = WMStepHelper(step).name()
        from WMTaskSpace import taskSpace
        stepSpace = taskSpace.stepSpace(stepName)

        os.chdir(stepSpace.location)

    def toTaskDirectory(self):
        """
        _toTaskDirectory_

        Switch to current working directory to the task location
        within WMTaskSpace

        """
        from WMTaskSpace import taskSpace
        os.chdir(taskSpace.location)
        return



#!/usr/bin/env python

"""
_Watchdog_

This cleverly named object is the thread that handles the monitoring of individual jobs
"""
from __future__ import division
from __future__ import print_function
from future.utils import viewitems

import logging
import os
import os.path
import threading
import traceback

from PSetTweaks.WMTweak import resizeResources
from WMCore.WMException import WMException
from WMCore.WMFactory import WMFactory


class WatchdogException(WMException):
    """
    _StepFactortyException_

    It's like an exception class that does nothing

    """
    pass


class Watchdog(threading.Thread):
    """
    Watchdog

    It's like a watchdog!  It watches things, and then it barks at them, curls up, and goes to sleep
    Or possibly monitors them and records what they do.  It's a bit of a crapshoot at this point.
    """

    def __init__(self, logPath=None, config=None):
        threading.Thread.__init__(self)
        self.doMonitoring = True
        self._Finished = threading.Event()
        self._EndOfJob = threading.Event()
        self._NewTask = threading.Event()
        self._JobKilled = threading.Event()
        self._RunUpdate = threading.Event()
        self._Interval = 120.0
        self._Monitors = []

        # Right now we join this, because we don't know
        # Where we'll be when we need this.
        self.logPath = os.path.join(os.getcwd(), logPath)

        self.factory = WMFactory(self.__class__.__name__,
                                 "WMCore.WMRuntime.Monitors")

    def setupMonitors(self, task, wmbsJob):
        logging.info("In Watchdog.setupMonitors")
        if not hasattr(task.data, 'watchdog'):
            msg = "Could not find watchdog in spec"
            logging.error(msg)
            # I don't think this is necessarily fatal
            return
        if not hasattr(task.data.watchdog, 'monitors'):
            msg = "Watchdog has no monitors"
            logging.error(msg)
            # Probably not fatal either
            return
        if hasattr(task.data.watchdog, 'interval'):
            # Set the interval off the config
            self.setInterval(task.data.watchdog.interval)
        for monitor in task.data.watchdog.monitors:
            msg = "Initializing monitor %s" % monitor
            logging.info(msg)
            mon = self.loadMonitor(monitor)
            args = {}
            if hasattr(task.data.watchdog, monitor):
                # This should be a config section
                monitorArgs = getattr(task.data.watchdog, monitor)
                args = monitorArgs.dictionary_()
            if monitor == 'PerformanceMonitor' and args:
                # Apply tweaks to PerformanceMonitor only.
                # Scale resources according to the HTCondor runtime environment.
                origCores = 1
                for stepName in task.listAllStepNames():
                    sh = task.getStepHelper(stepName)
                    origCores = max(origCores, sh.getNumberOfCores())
                resources = {'cores': origCores}
                origMaxPSS = args.get('maxPSS', args.get('maxRSS'))
                if origMaxPSS:
                    resources['memory'] = origMaxPSS
                # Actually parses the HTCondor runtime
                resizeResources(resources)
                # We decided to only touch Watchdog settings if the number of cores changed.
                # (even if this means the watchdog memory is wrong for a slot this size).
                changedCores = origCores != resources['cores']
                # If we did base maxPSS off the memory in the HTCondor slot, subtract a bit
                # off the top so watchdog triggers before HTCondor does.
                # Add the new number of cores to the args such that PerformanceMonitor can see it
                args['cores'] = resources['cores']
                if changedCores:
                    if origMaxPSS:
                        args['maxPSS'] = resources['memory'] - 50

                logging.info("Watchdog modified: %s. Final settings:", changedCores)
                for k, v in viewitems(args):
                    logging.info("  %s: %r", k, v)
            # Actually initialize the monitor variables
            mon.initMonitor(task=task, job=wmbsJob,
                            logPath=self.logPath, args=args)
            self._Monitors.append(mon)

        return

    def loadMonitor(self, monitorName):
        try:
            return self.factory.loadObject(monitorName)
        except WMException:
            msg = "WatchdogFactory Unable to load Object: %s" % monitorName
            logging.error(msg)
            raise WatchdogException(msg)
        except Exception as ex:
            msg = "Error creating object %s in WatchdogFactory:\n" % monitorName
            msg += str(ex)
            logging.error(msg)
            raise WatchdogException(msg)

    def setInterval(self, interval):
        """
        Set the monitor interval
        """
        logging.info("Set Watchdog interval to %s", interval)
        self._Interval = interval

    def disableMonitoring(self):
        """
        _disableMonitoring_

        Turn off active monitoring (periodicUpdate calls are disabled,
        event driven monitoring still occurs
        """
        self.doMonitoring = False

    def shutdown(self):
        """
        Shutdown the monitor.
        """
        logging.info("MonitorState: Shutdown called")
        self._Finished.set()
        return

    #  //=========notify Methods called by the ExecutionManager====
    # //
    # //  Start notification from the exe thread, this starts the
    #  //periodic updates of the monitor thread
    # //
    # //
    def notifyJobStart(self, task):
        """
        Start the job.
        """
        logging.info("MonitorThread: JobStarted")
        if self.doMonitoring:
            self.setDaemon(1)
            self.start()
        for monitor in self._Monitors:
            try:
                monitor.jobStart(task)
            except Exception as ex:
                msg = "Error in notifyJobStart for monitor class %s in Watchdog:\n" % monitor.__class__
                msg += str(ex)
                msg += str(traceback.format_exc())
                logging.error(msg)
                raise WatchdogException(msg)
        return

    #  //
    # // notify Monitors of new task start up
    # //
    def notifyStepStart(self, step):
        """
        notify Monitors of new task start up.
        """
        self._RunUpdate.set()
        for monitor in self._Monitors:
            try:
                monitor.stepStart(step)
            except Exception as ex:
                msg = "Error in notifyTaskStart for monitor class %s in Watchdog:\n" % monitor.__class__
                msg += str(ex)
                msg += str(traceback.format_exc())
                logging.error(msg)
                raise WatchdogException(msg)
        return

    #  //
    # // notify Monitors of task completion
    # //
    def notifyStepEnd(self, step, exitCode=0, stepReport=None):
        """
        notify Monitors of task completion.
        """
        self._RunUpdate.clear()
        for monitor in self._Monitors:
            try:
                monitor.stepEnd(step=step, stepReport=stepReport)
            except Exception as ex:
                msg = "Error in notifyTaskEnd for monitor class %s in Watchdog:\n" % monitor.__class__
                msg += str(ex)
                msg += str(traceback.format_exc())
                logging.error(msg)
                raise WatchdogException(msg)
        # print "Task Ended: %s with Exit Code:%s" % (task, exitCode)
        # self._MonMgr.taskEnd(task, exitCode)
        return

    #  //
    # // notify monitors of Job Completion, stops the periodic
    # //  updating
    def notifyJobEnd(self, task):
        """
        notify monitors of Job Completion, stops the periodic
        updating.
        """
        logging.info("MonitorThread: JobEnded")
        for monitor in self._Monitors:
            try:
                monitor.jobEnd(task)
            except Exception as ex:
                msg = "Error in notifyJobEnd for monitor class %s in Watchdog:\n" % monitor.__class__
                msg += str(ex)
                msg += str(traceback.format_exc())
                raise WatchdogException(msg)
        # self._MonMgr.jobEnd()
        self.shutdown()
        return

    #  //
    # //  Interrupt Notifiers
    # //   Job has been killed
    def notifyKillJob(self):
        """
        Interrupt Notifiers, Job has been killed.
        """
        logging.info("MonitorThread: JobKilled")
        for monitor in self._Monitors:
            try:
                monitor.jobKilled()
            except Exception as ex:
                msg = "Error in notifyKillJob for monitor class %s in Watchdog:\n" % monitor.__class__
                msg += str(ex)
                msg += str(traceback.format_exc())
                logging.error(msg)
                raise WatchdogException(msg)
        # self._MonMgr.jobKilled()
        self.shutdown()

    #  //
    # //  Task has been killed
    # //
    def notifyKillStep(self, step=None):
        """
        Task has been killed.
        """
        logging.info("MonitorThread: TaskKilled")
        for monitor in self._Monitors:
            try:
                monitor.stepKilled(step)
            except Exception as ex:
                msg = "Error in notifyKillTask for monitor class %s in Watchdog:\n" % monitor.__class__
                msg += str(ex)
                msg += str(traceback.format_exc())
                logging.error(msg)
                raise WatchdogException(msg)
                # self._MonMgr.taskKilled()

    # //
    # // Override Thread.run() to do the periodic update
    # //  of the MonitorState object and dispatch it to the monitors
    def run(self):
        """
        Override Thread.run() to do the periodic update
        of the MonitorState object and dispatch it to the monitors
        """
        while True:
            #  //
            # // shutdown signal
            # //
            if self._Finished.isSet():
                return

            # //
            # // Update State information only during a running task
            # //
            if self._RunUpdate.isSet():
                for monitor in self._Monitors:
                    try:
                        monitor.periodicUpdate()
                    except Exception as ex:
                        msg = "Error in periodicUpdate for monitor class %s in Watchdog:\n" % monitor.__class__
                        msg += str(ex)
                        msg += str(traceback.format_exc())
                        msg += "This is a CRITICAL error because this kills the monitoring.\n"
                        msg += "Terminate thread and retry.\n"
                        logging.error(msg)
                        # raise WatchdogException(msg)
                        # This one needs to be killed by itself
                        # since it's run by thread
                        os.abort()
                        # self._MonMgr.periodicUpdate()

            # time.sleep(self._Interval)
            self._Finished.wait(self._Interval)

    # //
    # // Load Monitor Objects based on Cfg settings passed
    # //  from Executor
    def initMonitorFwk(self, monitorCfg, updatorCfg):
        """
        _initMonitorFwk_

        Initialise the MonitorMgr object when this method is
        called from the Execution thread
        Load Monitor Objects based on Cfg settings passed from Executor.
        """
        # self._MonMgr.monitorConfig = monitorCfg
        # self._MonMgr.updatorConfig = updatorCfg
        # self._MonMgr.loadMonitors()
        # self._MonMgr.loadUpdators()
        return



#!/usr/bin/env python
"""
_CountPhedexNotUploaded_

MySQL implementation of DBSBuffer.CountPhedexNotUploaded

"""

from __future__ import division
from WMCore.Database.DBFormatter import DBFormatter


class CountPhedexNotUploaded(DBFormatter):
    """
    CountPhedexNotUploaded

    Count the number of files in dbs not uploaded to phedex, used in drain statistics

    """
    sql = """SELECT COUNT(*) FROM dbsbuffer_file
             WHERE in_phedex=0
             AND block_id IS NOT NULL
             AND lfn NOT LIKE :unmerged
             AND lfn NOT LIKE :mcfakefile
             AND lfn NOT LIKE :backfill
             AND lfn NOT LIKE :storeuser"""

    def execute(self, conn=None, transaction=False):
        binds = {'unmerged': '%unmerged%',
                 'mcfakefile': 'MCFakeFile%',
                 'backfill': '%BACKFILL%',
                 'storeuser': '/store/user%'}

        result = self.dbi.processData(self.sql, binds, conn=conn,
                                      transaction=transaction)
        result = self.format(result)
        return result[0][0]



#/usr/bin/env python
"""
_Destroy_

Clear out the ResourceControl schema.
"""




import threading
from WMCore.Database.DBCreator import DBCreator

class Destroy(DBCreator):
    def __init__(self, logger = None, dbi = None):
        """
        _init_

        Call the DBCreator constructor and create the list of required tables.
        """
        myThread = threading.currentThread()

        if logger == None:
            logger = myThread.logger
        if dbi == None:
            dbi = myThread.dbi

        DBCreator.__init__(self, logger, dbi)

        self.create = {}
        self.delete = {}
        self.delete["rc_threshold"] = "DROP TABLE rc_threshold"

        return



#!/usr/bin/env python
"""
_NewState_

MySQL implementation for labeling a job Complete
"""


from WMCore.Database.DBFormatter import DBFormatter

class NewState(DBFormatter):
    """
    _NewState_

    Insert new states into bl_status
    """


    sql = """INSERT IGNORE INTO bl_status (name) VALUES (:name)"""



    def execute(self, states, conn = None, transaction = False):
        """
        _execute_

        Complete jobs
        """

        binds = []
        if isinstance(states, str):
            binds = {'name': states}
        for state in states:
            binds.append({'name': state})


        result = self.dbi.processData(self.sql, binds, conn = conn,
                                      transaction = transaction)

        return



'''
Created on Jun 16, 2009

@author: meloam
'''
from __future__ import print_function
import os
import sys
from types import *

class ProcessMonitor(object):
    '''
    Lets us fork (and optionally exec) processes, monitoring their exit codes
    '''
    def __init__(self):
        '''
        Constructor
        '''
        self.processList = {}
        self.returnedList = {}

    def executeAndMonitor(self, child):
        pid = child.forkAndExecute()
        self.processList[pid] = child

    def checkChildren(self, deleteOldOnes = True):
        """
            Checks the status of all our children, updating our lists
        """
        if (deleteOldOnes):
            self.returnedList = {}

        for child in self.processList:
            newstatus, newsignal = child.isRunning()
            if (not ((newstatus == True) and (newsignal == True))):
                # if they BOTH aren't true, we have something to do
                self.returnedList[child.processID] = child
                del self.processList[child.processID]
                if (child.callback):
                    child.callback( child )



class ChildProcess(object):
    """
        base class for child processes
    """

    def __init__(self):
        self.processID = -1
        self.callback  = None
        self.ourStdout = None
        self.ourStderr = None

    def setStdout(self, handle):
        self.ourStdout = handle

    def setStderr(self, handle):
        self.ourStderr = handle

    def setCallback(self, newCallback):
        self.callback = newCallback

    def forkAndExecute(self):
        pid = os.fork()
        if (not pid):
            try:
                # we're in the child
                if (self.ourStderr):
                    sys.stderr = self.ourStderr
                if (self.ourStdout):
                    sys.stderr = self.ourStdout
                exitCode = self.execute()
                print("Falling through ChildProcess.forkAndExecute with code %s" %\
                         exitCode)
                sys.stdout.flush()
                sys.stderr.flush()
                os._exit( exitCode )
            except Exception as e:
                print("Something bad happened in ChildProcess.forkAndExecute in the child: %s" % e)
                os._exit(99)

        else:
            # we're in the parent
            self.processID = pid
            return pid

        raise RuntimeError("Something bad happened in fork()")

    def execute(self):
        """
            overridden in child classes to provide specific things to do
        """
        msg = "ProcessMonitor.ProcessMonitor.execute method not overridden in "
        msg += "implementation: %s\n" % self.__class__.__name__
        raise NotImplementedError(msg)

    def isRunning(self):
        if (self.processID == -1):
            raise RuntimeError("Trying to waitpid on nonexistant process")
        pid, status = os.waitpid( self.processID, os.WNOHANG )
        if ((pid == 0) and (status == 0)):
            # we're definately still running if this pops up
            return (True, True)
        else:
            # the child returned, get the status back from it
            # os.waitpid does dumb things to the exit code
            self.realstatus = ((0xFF00 & status) >> 8)
            self.realsignal = ((0x00FF & status))

            return (self.realstatus, self.realsignal)

class ExecProcess(ChildProcess):
    """
        calls a given external executable
    """

    def __init__(self):
        self.args = []

    def setArgs(self, arguments):
        """
            Either accepts a list of arguments which are passed to execvp
            OR     accepts a string which is passed to bash and shell-expanded
        """
        if not isinstance(arguments, list):
            # we got passed a string, pass it to a shell
            self.args[0] = 'bash'
            self.args[1] = '-c'
            self.args[2] = arguments
        else:
            # we got passed a list
            self.args = arguments

    def execute(self):
        if (self.args == []):
            raise RuntimeError("No arguments were set")
        os.execvp(self.args[0], self.args[1:])

class PythonProcess(ChildProcess):
    """
        Calls a function specified by the user
        to pass arguments, wrap it in a lambda
    """
    def __init__(self):
        self.target = None

    def setTarget(self, newtarget):
        if not isinstance(newtarget, (FunctionType, LambdaType)):
            raise RuntimeError("PythonProcess requires a function for target")

        self.target = newtarget

    def execute(self):
        if (self.target == None):
            raise RuntimeError("No execute process was set")
        return self.target()



#!/usr/bin/env python
"""
_BulkNewReturn_

Oracle implementation of JobGroup.BulkNewReturn
"""




from WMCore.WMBS.MySQL.JobGroup.BulkNewReturn import BulkNewReturn as MySQLBulkNewReturn

class BulkNewReturn(MySQLBulkNewReturn):
    """
    Does a bulk commit of jobGroups, followed by returning their IDs and UIDs

    """

    sql = """INSERT INTO wmbs_jobgroup (id, subscription, guid, output,
                      last_update) VALUES (wmbs_jobgroup_SEQ.nextval, :subscription,
                      :guid, :output, :timestamp)"""



#!/usr/bin/env python
"""
_StageOutMgr_

Util class to provide stage out functionality as an interface object.

Based of RuntimeStageOut.StageOutManager, that should probably eventually
use this class as a basic API
"""
from __future__ import print_function
from builtins import object
from future.utils import viewitems

import logging
# If we don't import them, they cannot be ever used (bad PyCharm!)
import WMCore.Storage.Backends
import WMCore.Storage.Plugins

from WMCore.Storage.DeleteMgr import DeleteMgr
from WMCore.Storage.Registry import retrieveStageOutImpl
from WMCore.Storage.StageOutError import StageOutFailure
from WMCore.Storage.StageOutError import StageOutInitError
from WMCore.WMException import WMException


def stageoutPolicyReport(fileToStage, pnn, command, stageOutType, stageOutExit):
    """
    Prepare some extra information regarding the stage out step (for both prod/analysis jobs).

    NOTE: this information used to be shipped to the old SSB dashboard. I'm unsure
    whether it's provided to any other monitoring system at the moment.
    """
    tempDict = {}
    tempDict['LFN'] = fileToStage['LFN'] if 'LFN' in fileToStage else None
    tempDict['PNN'] = fileToStage['PNN'] if 'PNN' in fileToStage else None
    tempDict['PNN'] = pnn if pnn else tempDict['PNN']
    tempDict['StageOutCommand'] = fileToStage['command'] if 'command' in fileToStage else None
    tempDict['StageOutCommand'] = command if command else tempDict['StageOutCommand']
    tempDict['StageOutType'] = stageOutType
    tempDict['StageOutExit'] = stageOutExit
    fileToStage['StageOutReport'].append(tempDict)
    return fileToStage


class StageOutMgr(object):
    """
    _StageOutMgr_

    Object that can be used to stage out a set of files
    using TFC or an override.

    """

    def __init__(self, **overrideParams):
        logging.info("StageOutMgr::__init__()")
        self.overrideConf = overrideParams

        # Figure out if any of the override parameters apply to stage-out
        self.override = False
        if overrideParams != {}:
            logging.info("StageOutMgr::__init__(): Override: %s", overrideParams)
            checkParams = ["command", "option", "phedex-node", "lfn-prefix"]
            for param in checkParams:
                if param in self.overrideConf:
                    self.override = True
            if not self.override:
                logging.info("=======StageOut Override: These are not the parameters you are looking for")

        self.substituteGUID = True
        self.fallbacks = []

        #  //
        # // Try an get the TFC for the site
        # //
        self.tfc = None

        self.numberOfRetries = 3
        self.retryPauseTime = 600

        from WMCore.Storage.SiteLocalConfig import loadSiteLocalConfig

        #  //
        # // If override isnt None, we dont need SiteCfg, if it is
        # //  then we need siteCfg otherwise we are dead.

        if self.override == False:
            self.siteCfg = loadSiteLocalConfig()

        if self.override:
            self.initialiseOverride()
        else:
            self.initialiseSiteConf()

        self.failed = {}
        self.completedFiles = {}
        return

    def initialiseSiteConf(self):
        """
        _initialiseSiteConf_

        Extract required information from site conf and TFC

        """

        implName = self.siteCfg.localStageOut.get("command", None)
        if implName == None:
            msg = "Unable to retrieve local stage out command\n"
            msg += "From site config file.\n"
            msg += "Unable to perform StageOut operation"
            raise StageOutInitError(msg)
        msg = "Local Stage Out Implementation to be used is: "
        msg += "%s\n" % implName

        pnn = self.siteCfg.localStageOut.get("phedex-node", None)
        if pnn == None:
            msg = "Unable to retrieve local stage out phedex-node\n"
            msg += "From site config file.\n"
            msg += "Unable to perform StageOut operation"
            raise StageOutInitError(msg)
        msg += "Local Stage Out PNN to be used is %s\n" % pnn
        catalog = self.siteCfg.localStageOut.get("catalog", None)
        if catalog == None:
            msg = "Unable to retrieve local stage out catalog\n"
            msg += "From site config file.\n"
            msg += "Unable to perform StageOut operation"
            raise StageOutInitError(msg)
        msg += "Local Stage Out Catalog to be used is %s\n" % catalog

        try:
            self.tfc = self.siteCfg.trivialFileCatalog()
            msg += "Trivial File Catalog has been loaded:\n"
            msg += str(self.tfc)
        except Exception as ex:
            msg = "Unable to load Trivial File Catalog:\n"
            msg += "Local stage out will not be attempted\n"
            msg += str(ex)
            raise StageOutInitError(msg)

        self.fallbacks = self.siteCfg.fallbackStageOut

        msg += "There are %s fallback stage out definitions.\n" % len(self.fallbacks)
        for item in self.fallbacks:
            pnn = item.get("phedex-node")
            command = item.get("command")
            if pnn is None:
                msg = "Unable to retrieve fallback phedex-node\n"
                msg += "From site config file.\n"
                msg += "Unable to perform StageOut operation"
                raise StageOutInitError(msg)
            if command is None:
                msg = "Unable to retrieve fallback command\n"
                msg += "From site config file.\n"
                msg += "Unable to perform StageOut operation"
                raise StageOutInitError(msg)
            msg += "\tFallback to : %s using: %s \n" % (pnn, command)

        logging.info("==== Stageout configuration start ====")
        logging.info(msg)
        logging.info("==== Stageout configuration finish ====")
        return

    def initialiseOverride(self):
        """
        _initialiseOverride_

        Extract and verify that the Override parameters are all present

        """
        overrideConf = self.overrideConf
        overrideParams = {
            "command": None,
            "option": None,
            "phedex-node": None,
            "lfn-prefix": None,
        }

        try:
            overrideParams['command'] = overrideConf['command']
            overrideParams['phedex-node'] = overrideConf['phedex-node']
            overrideParams['lfn-prefix'] = overrideConf['lfn-prefix']
        except Exception as ex:
            msg = "Unable to extract Override parameters from config:\n"
            msg += str(ex)
            raise StageOutInitError(msg)
        if 'option' in overrideConf:
            if len(overrideConf['option']) > 0:
                overrideParams['option'] = overrideConf['option']
            else:
                overrideParams['option'] = ""

        msg = "=======StageOut Override Initialised:================\n"
        for key, val in viewitems(overrideParams):
            msg += " %s : %s\n" % (key, val)
        msg += "=====================================================\n"
        logging.info(msg)
        self.fallbacks = []
        self.fallbacks.append(overrideParams)
        return

    def __call__(self, fileToStage):
        """
        _operator()_

        Use call to invoke transfers

        """
        lastException = Exception("empty exception")

        logging.info("==>Working on file: %s", fileToStage['LFN'])
        lfn = fileToStage['LFN']

        fileToStage['StageOutReport'] = []
        #  //
        # // No override => use local-stage-out from site conf
        # //  invoke for all files and check failures/successes
        if not self.override:
            logging.info("===> Attempting Local Stage Out.")
            try:
                pfn = self.localStageOut(lfn, fileToStage['PFN'], fileToStage.get('Checksums'))
                fileToStage['PFN'] = pfn
                fileToStage['PNN'] = self.siteCfg.localStageOut['phedex-node']
                fileToStage['StageOutCommand'] = self.siteCfg.localStageOut['command']
                self.completedFiles[fileToStage['LFN']] = fileToStage

                logging.info("===> Stage Out Successful: %s", fileToStage)
                fileToStage = stageoutPolicyReport(fileToStage, None, None, 'LOCAL', 0)
                return fileToStage
            except WMException as ex:
                lastException = ex
                logging.info("===> Local Stage Out Failure for file:")
                logging.info("======>  %s\n", fileToStage['LFN'])
                fileToStage = stageoutPolicyReport(fileToStage, self.siteCfg.localStageOut.get('phedex-node', None),
                                                   self.siteCfg.localStageOut['command'], 'LOCAL', 60311)
            except Exception as ex:
                lastException = StageOutFailure("Error during local stage out",
                                                error=str(ex))
                logging.info("===> Local Stage Out Failure for file:\n")
                logging.info("======>  %s\n", fileToStage['LFN'])
                fileToStage = stageoutPolicyReport(fileToStage, self.siteCfg.localStageOut.get('phedex-node', None),
                                                   self.siteCfg.localStageOut['command'], 'LOCAL', 60311)

        # //
        # // Still here => failure, start using the fallback stage outs
        # //  If override is set, then that will be the only fallback available
        logging.info("===> Attempting %s Fallback Stage Outs", len(self.fallbacks))
        for fallback in self.fallbacks:
            try:
                pfn = self.fallbackStageOut(lfn, fileToStage['PFN'],
                                            fallback, fileToStage.get('Checksums'))
                fileToStage['PFN'] = pfn
                fileToStage['PNN'] = fallback['phedex-node']
                fileToStage['StageOutCommand'] = fallback['command']
                logging.info("attempting fallback")
                self.completedFiles[fileToStage['LFN']] = fileToStage
                if lfn in self.failed:
                    del self.failed[lfn]

                logging.info("===> Stage Out Successful: %s", fileToStage)
                fileToStage = stageoutPolicyReport(fileToStage, None, None, 'FALLBACK', 0)
                return fileToStage
            except Exception as ex:
                fileToStage = stageoutPolicyReport(fileToStage, fallback.get('phedex-node', None),
                                                   fallback['command'], 'FALLBACK', 60310)
                lastException = ex
                continue

        raise lastException

    def fallbackStageOut(self, lfn, localPfn, fbParams, checksums):
        """
        _fallbackStageOut_

        Given the lfn and parameters for a fallback stage out, invoke it

        parameters should contain:

        command - the stage out impl plugin name to be used
        option - the option values to be passed to that command (None is allowed)
        lfn-prefix - the LFN prefix to generate the PFN
        phedex-node - the Name of the PNN to which the file is being xferred

        """
        pfn = "%s%s" % (fbParams['lfn-prefix'], lfn)

        try:
            impl = retrieveStageOutImpl(fbParams['command'])
        except Exception as ex:
            msg = "Unable to retrieve impl for fallback stage out:\n"
            msg += "Error retrieving StageOutImpl for command named: "
            msg += "%s\n" % fbParams['command']
            raise StageOutFailure(msg, Command=fbParams['command'],
                                  LFN=lfn, ExceptionDetail=str(ex))

        impl.numRetries = self.numberOfRetries
        impl.retryPause = self.retryPauseTime

        try:
            impl(fbParams['command'], localPfn, pfn, fbParams.get("option", None), checksums)
        except Exception as ex:
            msg = "Failure for fallback stage out:\n"
            msg += str(ex)
            raise StageOutFailure(msg, Command=fbParams['command'],
                                  LFN=lfn, InputPFN=localPfn, TargetPFN=pfn)

        return pfn

    def localStageOut(self, lfn, localPfn, checksums):
        """
        _localStageOut_

        Given the lfn and local stage out params, invoke the local stage out

        """
        command = self.siteCfg.localStageOut['command']
        options = self.siteCfg.localStageOut.get('option', None)
        pfn = self.searchTFC(lfn)
        protocol = self.tfc.preferredProtocol
        if pfn == None:
            msg = "Unable to match lfn to pfn: \n  %s" % lfn
            raise StageOutFailure(msg, LFN=lfn, TFC=str(self.tfc))

        try:
            impl = retrieveStageOutImpl(command)
        except Exception as ex:
            msg = "Unable to retrieve impl for local stage out:\n"
            msg += "Error retrieving StageOutImpl for command named: %s\n" % (
                command,)
            raise StageOutFailure(msg, Command=command,
                                  LFN=lfn, ExceptionDetail=str(ex))
        impl.numRetries = self.numberOfRetries
        impl.retryPause = self.retryPauseTime

        try:
            impl(protocol, localPfn, pfn, options, checksums)
        except Exception as ex:
            msg = "Failure for local stage out:\n"
            msg += str(ex)
            try:
                import traceback
                msg += traceback.format_exc()
            except AttributeError as ex:
                msg += "Traceback unavailable\n"
            raise StageOutFailure(msg, Command=command, Protocol=protocol,
                                  LFN=lfn, InputPFN=localPfn, TargetPFN=pfn)

        return pfn

    def cleanSuccessfulStageOuts(self):
        """
        _cleanSucessfulStageOuts_

        In the event of a failed stage out, this method can be called to cleanup the
        files that may have previously been staged out so that the job ends in a clear state
        of failure, rather than a partial success


        """
        for lfn, fileInfo in viewitems(self.completedFiles):
            pfn = fileInfo['PFN']
            command = fileInfo['StageOutCommand']
            msg = "Cleaning out file: %s\n" % lfn
            msg += "Removing PFN: %s" % pfn
            msg += "Using command implementation: %s\n" % command
            logging.info(msg)
            delManager = DeleteMgr(**self.overrideConf)
            try:
                delManager.deletePFN(pfn, lfn, command)
            except StageOutFailure as ex:
                msg = "Failed to cleanup staged out file after error:"
                msg += " %s\n%s" % (lfn, str(ex))
                logging.error(msg)

    def searchTFC(self, lfn):
        """
        _searchTFC_

        Search the Trivial File Catalog for the lfn provided,
        if a match is made, return the matched PFN

        """
        if self.tfc == None:
            msg = "Trivial File Catalog not available to match LFN:\n"
            msg += lfn
            logging.error(msg)
            return None
        if self.tfc.preferredProtocol == None:
            msg = "Trivial File Catalog does not have a preferred protocol\n"
            msg += "which prevents local stage out for:\n"
            msg += lfn
            logging.error(msg)
            return None

        pfn = self.tfc.matchLFN(self.tfc.preferredProtocol, lfn)
        if pfn == None:
            msg = "Unable to map LFN to PFN:\n"
            msg += "LFN: %s\n" % lfn
            return None

        msg = "LFN to PFN match made:\n"
        msg += "LFN: %s\nPFN: %s\n" % (lfn, pfn)
        logging.info(msg)
        return pfn



#!/usr/bin/env python
"""
WorkQueue PolicyInterface

"""
from builtins import object
import logging
from copy import deepcopy




class PolicyInterface(object):
    """Interface for policies"""
    def __init__(self, **args):
        self.logger = args.pop('logger') if args.get('logger') else logging.getLogger()
        self.args = deepcopy(args)



#!/usr/bin/env python

"""
This is a direct copy of code in cmssw/FWCore/PythonUtilities/python/LumiList.py
It can be removed if we ever have a core set of code shared between DMWM and CMSSW

Handle lists of lumi sections. Constuct in several different formats and filter
(mask) a secondary list of lumis.
This class can also handle ranges of events as the structure is identical
or could be subclassed renaming a function or two.

This code began life in COMP/CRAB/python/LumiList.py
"""


from builtins import zip, range, next, str, object

from future.utils import viewitems

from future import standard_library
standard_library.install_aliases()

import copy
import itertools
import json
import re
import urllib.request
from contextlib import closing

class LumiList(object):
    """
    Deal with lists of lumis in several different forms:
    Compact list:
        {
        '1': [[1, 33], [35, 35], [37, 47], [49, 75], [77, 130], [133, 136]],
        '2':[[1,45],[50,80]]
        }
        where the first key is the run number, subsequent pairs are
        ranges of lumis within that run that are desired
    Runs and lumis:
        {
        '1': [1,2,3,4,6,7,8,9,10],
        '2': [1,4,5,20]
        }
        where the first key is the run number and the list is a list of
        individual lumi sections. This form also takes a list of these objects
        which can be much faster than LumiList += LumiList
    Run  lumi pairs:
        [[1,1], [1,2],[1,4], [2,1], [2,5], [1,10]]
        where each pair in the list is an individual run&lumi
    CMSSW representation:
        '1:1-1:33,1:35,1:37-1:47,2:1-2:45,2:50-2:80'
        The string used by CMSSW in lumisToProcess or lumisToSkip
        is a subset of the compactList example above
    """

    def __init__(self, filename=None, lumis=None, runsAndLumis=None, runs=None, compactList=None, url=None,
                 wmagentFormat=None):
        """
        Constructor takes filename (JSON), a list of run/lumi pairs,
        or a dict with run #'s as the keys and a list of lumis as the values, or just a list of runs
        """
        self.compactList = {}
        self.duplicates = {}
        if filename:
            self.filename = filename
            with open(self.filename,'r') as jsonFile:
                self.compactList = json.load(jsonFile)
        elif url:
            self.url = url
            with closing(urllib.request.urlopen(url)) as jsonFile:
                self.compactList = json.load(jsonFile)
        elif lumis:
            runsAndLumis = {}
            for (run, lumi) in lumis:
                run = str(run)
                if run not in runsAndLumis:
                    runsAndLumis[run] = []
                runsAndLumis[run].append(lumi)

        if isinstance(runsAndLumis, list):
            queued = {}
            for runLumiList in runsAndLumis:
                for run, lumis in viewitems(runLumiList):
                    queued.setdefault(run, []).extend(lumis)
            runsAndLumis = queued

        if runsAndLumis:
            for run in runsAndLumis:
                runString = str(run)
                lastLumi = -1000
                lumiList = runsAndLumis[run]
                if lumiList:
                    self.compactList[runString] = []
                    self.duplicates[runString] = []
                    for lumi in sorted(int(l) for l in lumiList):
                        if lumi == lastLumi:
                            self.duplicates[runString].append(lumi)
                        elif lumi != lastLumi + 1: # Break in lumi sequence
                            self.compactList[runString].append([lumi, lumi])
                        else:
                            nRange =  len(self.compactList[runString])
                            self.compactList[runString][nRange-1][1] = lumi
                        lastLumi = lumi
        if runs:
            for run in runs:
                runString = str(run)
                self.compactList[runString] = [[1, 0xFFFFFFF]]

        if compactList:
            for run in compactList:
                runString = str(run)
                if compactList[run]:
                    self.compactList[runString] = compactList[run]

        if wmagentFormat:
            """
            Expecting a list or tuple of two elements. First is a list of run #s, the second is a list of strings.
            Each string is a comma separated list of lumi numbers. Even numbers are the start of a lumi range,
            odd numbers are the end of that lumi range.
            """

            if not isinstance(wmagentFormat, (list, tuple)) or len(wmagentFormat) != 2:
                raise RuntimeError('Improper format for wmagentFormat. Must be list or tuple of lists')

            runs, lumis = wmagentFormat

            if len(runs) <= 0 or len(lumis) != len(runs):
                raise RuntimeError('Improper format for wmagentFormat. # of lumi lists must match # of runs')

            for run, lumiString in zip(runs, lumis):
                runLumis = lumiString.split(',')
                if str(run) not in self.compactList:
                    self.compactList[str(run)] = []
                if len(runLumis) % 2:
                    raise RuntimeError('Improper format for wmagentFormat. Lumis must be in pairs')

                # Walk through the list two at a time, get both values
                it = iter(runLumis)
                for beginLumi in it:
                    endLumi = next(it)
                    self.compactList[str(run)].append([int(beginLumi), int(endLumi)])

        # Compact each run and make it unique

        for run in self.compactList:
            newLumis = []
            for lumi in sorted(self.compactList[run]):
                # If the next lumi starts inside or just after the last just change the endpoint of the first
                if newLumis and newLumis[-1][0] <= lumi[0] <= newLumis[-1][1] + 1:
                    newLumis[-1][1] = max(newLumis[-1][1], lumi[1])
                else:
                    newLumis.append(lumi)
            self.compactList[run] = newLumis

    def __sub__(self, other): # Things from self not in other
        result = {}
        for run in sorted(self.compactList.keys()):
            alumis = sorted(self.compactList[run])
            blumis = sorted(other.compactList.get(run, []))
            alist = []                    # verified part
            for alumi in alumis:
                tmplist = [alumi[0], alumi[1]] # may be part
                for blumi in blumis:
                    if blumi[0] <= tmplist[0] and blumi[1] >= tmplist[1]: # blumi has all of alumi
                        tmplist = []
                        break # blumi is has all of alumi
                    if blumi[0] > tmplist[0] and blumi[1] < tmplist[1]: # blumi is part of alumi
                        alist.append([tmplist[0], blumi[0]-1])
                        tmplist = [blumi[1]+1, tmplist[1]]
                    elif blumi[0] <= tmplist[0] and blumi[1] < tmplist[1] and blumi[1]>=tmplist[0]: # overlaps start
                        tmplist = [blumi[1]+1, tmplist[1]]
                    elif blumi[0] > tmplist[0] and blumi[1] >= tmplist[1] and blumi[0]<=tmplist[1]: # overlaps end
                        alist.append([tmplist[0], blumi[0]-1])
                        tmplist = []
                        break
                if tmplist:
                    alist.append(tmplist)
            result[run] = alist

        return LumiList(compactList = result)


    def __and__(self, other): # Things in both
        result = {}
        aruns = set(self.compactList.keys())
        bruns = set(other.compactList.keys())
        for run in aruns & bruns:
            lumiList = []                    # List for this run
            unique = []                    # List for this run
            for alumi in self.compactList[run]:
                for blumi in other.compactList[run]:
                    if blumi[0] <= alumi[0] and blumi[1] >= alumi[1]: # blumi has all of alumi
                        lumiList.append(alumi)
                    if blumi[0] > alumi[0] and blumi[1] < alumi[1]: # blumi is part of alumi
                        lumiList.append(blumi)
                    elif blumi[0] <= alumi[0] and blumi[1] < alumi[1] and blumi[1] >= alumi[0]: # overlaps start
                        lumiList.append([alumi[0], blumi[1]])
                    elif blumi[0] > alumi[0] and blumi[1] >= alumi[1] and blumi[0] <= alumi[1]: # overlaps end
                        lumiList.append([blumi[0], alumi[1]])


            if lumiList:
                unique = [lumiList[0]]
            for pair in lumiList[1:]:
                if pair[0] == unique[-1][1]+1:
                    unique[-1][1] = copy.deepcopy(pair[1])
                else:
                    unique.append(copy.deepcopy(pair))

            result[run] = unique
        return LumiList(compactList = result)


    def __or__(self, other):
        result = {}
        aruns = list(self.compactList)
        bruns = list(other.compactList)
        runs = set(aruns + bruns)
        for run in runs:
            overlap = sorted(self.compactList.get(run, []) + other.compactList.get(run, []))
            unique = [overlap[0]]
            for pair in overlap[1:]:
                if pair[0] >= unique[-1][0] and pair[0] <= unique[-1][1]+1 and pair[1] > unique[-1][1]:
                    unique[-1][1] = copy.deepcopy(pair[1])
                elif pair[0] > unique[-1][1]:
                    unique.append(copy.deepcopy(pair))
            result[run] = unique
        return LumiList(compactList = result)


    def __add__(self, other):
        # + is the same as |
        return self.__or__(other)

    def __len__(self):
        '''Returns number of runs in list'''
        return len(self.compactList)

    def filterLumis(self, lumiList):
        """
        Return a list of lumis that are in compactList.
        lumilist is of the simple form
        [(run1,lumi1),(run1,lumi2),(run2,lumi1)]
        """
        filteredList = []
        for (run, lumi) in lumiList:
            runsInLumi = self.compactList.get(str(run), [[0, -1]])
            for (first, last) in runsInLumi:
                if lumi >= first and lumi <= last:
                    filteredList.append((run, lumi))
                    break
        return filteredList


    def __str__ (self):
        doubleBracketRE = re.compile (r']],')
        return doubleBracketRE.sub (']],\n',
                                    json.dumps (self.compactList,
                                                sort_keys=True))

    def getCompactList(self):
        """
        Return the compact list representation
        """
        return self.compactList


    def getDuplicates(self):
        """
        Return the list of duplicates found during construction as a LumiList
        """
        return LumiList(runsAndLumis = self.duplicates)


    def getLumis(self):
        """
        Return the list of pairs representation
        """
        theList = []
        runs = list(self.compactList)
        runs.sort(key=int)
        for run in runs:
            lumis = self.compactList[run]
            for lumiPair in sorted(lumis):
                for lumi in range(lumiPair[0], lumiPair[1]+1):
                    theList.append((int(run), lumi))

        return theList


    def getRuns(self):
        '''
        return the sorted list of runs contained
        '''
        return sorted (self.compactList.keys())


    def _getLumiParts(self):
        """
        Turn compactList into a list of the format
        [ 'R1:L1', 'R2:L2-R2:L3' ] which is used by getCMSSWString and getVLuminosityBlockRange
        """

        parts = []
        runs = list(self.compactList)
        runs.sort(key=int)
        for run in runs:
            lumis = self.compactList[run]
            for lumiPair in sorted(lumis):
                if lumiPair[0] == lumiPair[1]:
                    parts.append("%s:%s" % (run, lumiPair[0]))
                else:
                    parts.append("%s:%s-%s:%s" %
                                 (run, lumiPair[0], run, lumiPair[1]))
        return parts


    def getCMSSWString(self):
        """
        Turn compactList into a list of the format
        R1:L1,R2:L2-R2:L3 which is acceptable to CMSSW LumiBlockRange variable
        """

        parts = self._getLumiParts()
        output = ','.join(parts)
        return str(output)


    def getVLuminosityBlockRange(self, tracked = False):
        """
        Turn compactList into an (optionally tracked) VLuminosityBlockRange
        """

        import FWCore.ParameterSet.Config as cms
        parts = self._getLumiParts()
        if tracked:
            return cms.VLuminosityBlockRange(parts)
        else:
            return cms.untracked.VLuminosityBlockRange(parts)


    def writeJSON(self, fileName):
        """
        Write out a JSON file representation of the object
        """
        with open(fileName,'w') as jsonFile:
            jsonFile.write("%s\n" % self)


    def removeRuns (self, runList):
        '''
        removes runs from runList from collection
        '''
        for run in runList:
            run = str(run)
            if run in self.compactList:
                del self.compactList[run]

        return


    def selectRuns (self, runList):
        '''
        Selects only runs from runList in collection
        '''
        runsToDelete = []
        for run in list(self.compactList):
            if int(run) not in runList and run not in runList:
                runsToDelete.append(run)

        for run in runsToDelete:
            del self.compactList[run]

        return

    def contains (self, run, lumiSection = None):
        '''
        returns true if the run, lumi section passed in is contained
        in this lumiList.  Input can be either:
        - a single tuple of (run, lumi),
        - separate run and lumi numbers
        - a single run number (returns true if any lumi sections exist)
        '''
        if lumiSection is None:
            # if this is an integer or a string, see if the run exists
            if isinstance (run, int) or isinstance (run, str):
                return str(run) in self.compactList
            # if we're here, then run better be a tuple or list
            try:
                lumiSection = run[1]
                run         = run[0]
            except:
                raise RuntimeError("Improper format for run '%s'" % run)
        lumiRangeList = self.compactList.get( str(run) )
        if not lumiRangeList:
            # the run isn't there, so no need to look any further
            return False
        for lumiRange in lumiRangeList:
            # we want to make this as found if either the lumiSection
            # is inside the range OR if the lumi section is greater
            # than or equal to the lower bound of the lumi range and
            # the upper bound is 0 (which means extends to the end of
            # the run)
            if lumiRange[0] <= lumiSection and \
               (0 == lumiRange[1] or lumiSection <= lumiRange[1]):
                # got it
                return True
        return False


    def __contains__ (self, runTuple):
        return self.contains (runTuple)



'''
# Unit test code
import unittest
import FWCore.ParameterSet.Config as cms

class LumiListTest(unittest.TestCase):
    """
    _LumiListTest_

    """

    def testRead(self):
        """
        Test reading from JSON
        """
        exString = "1:1-1:33,1:35,1:37-1:47,2:49-2:75,2:77-2:130,2:133-2:136"
        exDict   = {'1': [[1, 33], [35, 35], [37, 47]],
                    '2': [[49, 75], [77, 130], [133, 136]]}
        exVLBR   = cms.VLuminosityBlockRange('1:1-1:33', '1:35', '1:37-1:47', '2:49-2:75', '2:77-2:130', '2:133-2:136')

        jsonList = LumiList(filename = 'lumiTest.json')
        lumiString = jsonList.getCMSSWString()
        lumiList = jsonList.getCompactList()
        lumiVLBR = jsonList.getVLuminosityBlockRange(True)

        self.assertTrue(lumiString == exString)
        self.assertTrue(lumiList   == exDict)
        self.assertTrue(lumiVLBR   == exVLBR)

    def testList(self):
        """
        Test constucting from list of pairs
        """

        listLs1 = range(1, 34) + [35] + range(37, 48)
        listLs2 = range(49, 76) + range(77, 131) + range(133, 137)
        lumis = zip([1]*100, listLs1) + zip([2]*100, listLs2)

        jsonLister = LumiList(filename = 'lumiTest.json')
        jsonString = jsonLister.getCMSSWString()
        jsonList = jsonLister.getCompactList()

        pairLister = LumiList(lumis = lumis)
        pairString = pairLister.getCMSSWString()
        pairList = pairLister.getCompactList()

        self.assertTrue(jsonString == pairString)
        self.assertTrue(jsonList   == pairList)


    def testRuns(self):
        """
        Test constucting from run and list of lumis
        """
        runsAndLumis = {
            1: range(1, 34) + [35] + range(37, 48),
            2: range(49, 76) + range(77, 131) + range(133, 137)
        }
        runsAndLumis2 = {
            '1': range(1, 34) + [35] + range(37, 48),
            '2': range(49, 76) + range(77, 131) + range(133, 137)
        }
        blank = {
            '1': [],
            '2': []
        }

        jsonLister = LumiList(filename = 'lumiTest.json')
        jsonString = jsonLister.getCMSSWString()
        jsonList   = jsonLister.getCompactList()

        runLister = LumiList(runsAndLumis = runsAndLumis)
        runString = runLister.getCMSSWString()
        runList   = runLister.getCompactList()

        runLister2 = LumiList(runsAndLumis = runsAndLumis2)
        runList2 = runLister2.getCompactList()

        runLister3 = LumiList(runsAndLumis = blank)


        self.assertTrue(jsonString == runString)
        self.assertTrue(jsonList   == runList)
        self.assertTrue(runList2   == runList)
        self.assertTrue(len(runLister3) == 0)

    def testFilter(self):
        """
        Test filtering of a list of lumis
        """
        runsAndLumis = {
            1: range(1, 34) + [35] + range(37, 48),
            2: range(49, 76) + range(77, 131) + range(133, 137)
        }

        completeList = zip([1]*150, range(1, 150)) + \
                       zip([2]*150, range(1, 150)) + \
                       zip([3]*150, range(1, 150))

        smallList    = zip([1]*50,  range(1, 10)) + zip([2]*50, range(50, 70))
        overlapList  = zip([1]*150, range(30, 40)) + \
                       zip([2]*150, range(60, 80))
        overlapRes   = zip([1]*9,   range(30, 34)) + [(1, 35)] + \
                       zip([1]*9,   range(37, 40)) + \
                       zip([2]*30,  range(60, 76)) + \
                       zip([2]*9,   range(77, 80))

        runLister = LumiList(runsAndLumis = runsAndLumis)

        # Test a list to be filtered which is a superset of constructed list
        filterComplete = runLister.filterLumis(completeList)
        # Test a list to be filtered which is a subset of constructed list
        filterSmall    = runLister.filterLumis(smallList)
        # Test a list to be filtered which is neither
        filterOverlap  = runLister.filterLumis(overlapList)

        self.assertTrue(filterComplete == runLister.getLumis())
        self.assertTrue(filterSmall    == smallList)
        self.assertTrue(filterOverlap  == overlapRes)

    def testDuplicates(self):
        """
        Test a list with lots of duplicates
        """
        result = zip([1]*100, range(1, 34) + range(37, 48))
        lumis  = zip([1]*100, range(1, 34) + range(37, 48) + range(5, 25))

        lister = LumiList(lumis = lumis)
        self.assertTrue(lister.getLumis() == result)

    def testNull(self):
        """
        Test a null list
        """

        runLister = LumiList(lumis = None)

        self.assertTrue(runLister.getCMSSWString() == '')
        self.assertTrue(runLister.getLumis() == [])
        self.assertTrue(runLister.getCompactList() == {})

    def testSubtract(self):
        """
        a-b for lots of cases
        """

        alumis = {'1' : range(2,20) + range(31,39) + range(45,49),
                  '2' : range(6,20) + range (30,40),
                  '3' : range(10,20) + range (30,40) + range(50,60),
                 }
        blumis = {'1' : range(1,6) + range(12,13) + range(16,30) + range(40,50) + range(33,36),
                  '2' : range(10,35),
                  '3' : range(10,15) + range(35,40) + range(45,51) + range(59,70),
                 }
        clumis = {'1' : range(1,6) + range(12,13) + range(16,30) + range(40,50) + range(33,36),
                  '2' : range(10,35),
                 }
        result = {'1' : range(6,12) + range(13,16) + range(31,33) + range(36,39),
                  '2' : range(6,10) + range(35,40),
                  '3' : range(15,20) + range(30,35) + range(51,59),
                 }
        result2 = {'1' : range(6,12) + range(13,16) + range(31,33) + range(36,39),
                   '2' : range(6,10) + range(35,40),
                   '3' : range(10,20) + range (30,40) + range(50,60),
                 }
        a = LumiList(runsAndLumis = alumis)
        b = LumiList(runsAndLumis = blumis)
        c = LumiList(runsAndLumis = clumis)
        r = LumiList(runsAndLumis = result)
        r2 = LumiList(runsAndLumis = result2)

        self.assertTrue((a-b).getCMSSWString() == r.getCMSSWString())
        self.assertTrue((a-b).getCMSSWString() != (b-a).getCMSSWString())
        # Test where c is missing runs from a
        self.assertTrue((a-c).getCMSSWString() == r2.getCMSSWString())
        self.assertTrue((a-c).getCMSSWString() != (c-a).getCMSSWString())
        # Test empty lists
        self.assertTrue(str(a-a) == '{}')
        self.assertTrue(len(a-a) == 0)

    def testOr(self):
        """
        a|b for lots of cases
        """

        alumis = {'1' : range(2,20) + range(31,39) + range(45,49),
                  '2' : range(6,20) + range (30,40),
                  '3' : range(10,20) + range (30,40) + range(50,60),
                 }
        blumis = {'1' : range(1,6) + range(12,13) + range(16,30) + range(40,50) + range(39,80),
                  '2' : range(10,35),
                  '3' : range(10,15) + range(35,40) + range(45,51) + range(59,70),
                 }
        clumis = {'1' : range(1,6) + range(12,13) + range(16,30) + range(40,50) + range(39,80),
                  '2' : range(10,35),
                 }
        result = {'1' : range(2,20) + range(31,39) + range(45,49) + range(1,6) + range(12,13) + range(16,30) + range(40,50) + range(39,80),
                  '2' : range(6,20) + range (30,40) + range(10,35),
                  '3' : range(10,20) + range (30,40) + range(50,60) + range(10,15) + range(35,40) + range(45,51) + range(59,70),
                 }
        a = LumiList(runsAndLumis = alumis)
        b = LumiList(runsAndLumis = blumis)
        c = LumiList(runsAndLumis = blumis)
        r = LumiList(runsAndLumis = result)
        self.assertTrue((a|b).getCMSSWString() == r.getCMSSWString())
        self.assertTrue((a|b).getCMSSWString() == (b|a).getCMSSWString())
        self.assertTrue((a|b).getCMSSWString() == (a+b).getCMSSWString())

        # Test list constuction (faster)

        multiple = [alumis, blumis, clumis]
        easy = LumiList(runsAndLumis = multiple)
        hard = a + b
        hard += c
        self.assertTrue(hard.getCMSSWString() == easy.getCMSSWString())

    def testAnd(self):
        """
        a&b for lots of cases
        """

        alumis = {'1' : range(2,20) + range(31,39) + range(45,49),
                  '2' : range(6,20) + range (30,40),
                  '3' : range(10,20) + range (30,40) + range(50,60),
                  '4' : range(1,100),
                 }
        blumis = {'1' : range(1,6) + range(12,13) + range(16,25) + range(25,40) + range(40,50) + range(33,36),
                  '2' : range(10,35),
                  '3' : range(10,15) + range(35,40) + range(45,51) + range(59,70),
                  '5' : range(1,100),
                 }
        result = {'1' : range(2,6) + range(12,13) + range(16,20) + range(31,39) + range(45,49),
                  '2' : range(10,20) + range(30,35),
                  '3' : range(10,15) + range(35,40) + range(50,51)+ range(59,60),
                 }
        a = LumiList(runsAndLumis = alumis)
        b = LumiList(runsAndLumis = blumis)
        r = LumiList(runsAndLumis = result)
        self.assertTrue((a&b).getCMSSWString() == r.getCMSSWString())
        self.assertTrue((a&b).getCMSSWString() == (b&a).getCMSSWString())
        self.assertTrue((a|b).getCMSSWString() != r.getCMSSWString())

    def testRemoveSelect(self):
        """
        a-b for lots of cases
        """

        alumis = {'1' : range(2,20) + range(31,39) + range(45,49),
                  '2' : range(6,20) + range (30,40),
                  '3' : range(10,20) + range (30,40) + range(50,60),
                  '4' : range(10,20) + range (30,80),
                 }

        result = {'2' : range(6,20) + range (30,40),
                  '4' : range(10,20) + range (30,80),
                 }

        rem = LumiList(runsAndLumis = alumis)
        sel = LumiList(runsAndLumis = alumis)
        res = LumiList(runsAndLumis = result)

        rem.removeRuns([1,3])
        sel.selectRuns([2,4])

        self.assertTrue(rem.getCMSSWString() == res.getCMSSWString())
        self.assertTrue(sel.getCMSSWString() == res.getCMSSWString())
        self.assertTrue(sel.getCMSSWString() == rem.getCMSSWString())

    def testURL(self):
        URL = 'https://cms-service-dqm.web.cern.ch/cms-service-dqm/CAF/certification/Collisions12/8TeV/Reprocessing/Cert_190456-195530_8TeV_08Jun2012ReReco_Collisions12_JSON.txt'
        ll = LumiList(url=URL)
        self.assertTrue(len(ll) > 0)


    def testWrite(self):
        alumis = {'1' : range(2,20) + range(31,39) + range(45,49),
                  '2' : range(6,20) + range (30,40),
                  '3' : range(10,20) + range (30,40) + range(50,60),
                  '4' : range(1,100),
                 }
        a = LumiList(runsAndLumis = alumis)
        a.writeJSON('newFile.json')


if __name__ == '__main__':
    with open('lumiTest.json','w') as jsonFile:
        jsonFile.write('{"1": [[1, 33], [35, 35], [37, 47]], "2": [[49, 75], [77, 130], [133, 136]]}')
    unittest.main()
'''
# Test JSON file

#{"1": [[1, 33], [35, 35], [37, 47]], "2": [[49, 75], [77, 130], [133, 136]]}

if __name__ == '__main__':
    #############################################
    ## Load and save command line history when ##
    ## running interactively.                  ##
    #############################################
    import os, readline
    import atexit
    historyPath = os.path.expanduser("~/.pyhistory")


    def save_history(historyPath=historyPath):
        import readline
        readline.write_history_file(historyPath)
        if os.path.exists(historyPath):
            readline.read_history_file(historyPath)


    atexit.register(save_history)
    readline.parse_and_bind("set show-all-if-ambiguous on")
    readline.parse_and_bind("tab: complete")
    if os.path.exists (historyPath) :
        readline.read_history_file(historyPath)
        readline.set_history_length(-1)





"""
_Create_

Implementation of Create for Oracle.

Inherit from CreateWMBSBase, and add Oracle specific creates to the dictionary
at some high value.

Remove Oracle reserved words (e.g. size, file) and revise SQL used (e.g. no BOOLEAN)
"""

from WMCore.JobStateMachine.Transitions import Transitions
from WMCore.WMBS.CreateWMBSBase import CreateWMBSBase


class Create(CreateWMBSBase):
    """
    Class to set up the WMBS schema in an Oracle database
    """
    sequence_tables = [
        'wmbs_fileset',
        'wmbs_file_details',
        'wmbs_location_state',
        'wmbs_location',
        'wmbs_pnns',
        'wmbs_workflow',
        'wmbs_subscription',
        'wmbs_jobgroup',
        'wmbs_job',
        'wmbs_job_state',
        'wmbs_checksum_type',
        'wmbs_sub_types',
        'wmbs_users',
        'wmbs_workunit',
    ]

    def __init__(self, logger=None, dbi=None, params=None):
        """
        _init_

        Call the base class's constructor and create all necessary tables,
        constraints and inserts.
        """
        CreateWMBSBase.__init__(self, logger, dbi)

        tablespaceTable = ""
        tablespaceIndex = ""

        if params:
            if "tablespace_table" in params:
                tablespaceTable = "TABLESPACE %s" % params["tablespace_table"]
            if "tablespace_index" in params:
                tablespaceIndex = "USING INDEX TABLESPACE %s" % params["tablespace_index"]

        self.create["01wmbs_fileset"] = \
            """CREATE TABLE wmbs_fileset (
                 id          INTEGER      NOT NULL,
                 name        VARCHAR(1250) NOT NULL,
                 open        CHAR(1)      CHECK (open IN ('0', '1' )) NOT NULL,
                 last_update INTEGER      NOT NULL
                 ) %s""" % tablespaceTable

        self.indexes["01_pk_wmbs_fileset"] = \
            """ALTER TABLE wmbs_fileset ADD
                 (CONSTRAINT wmbs_fileset_pk PRIMARY KEY (id) %s)""" % tablespaceIndex

        self.indexes["02_pk_wmbs_fileset"] = \
            """ALTER TABLE wmbs_fileset ADD
                 (CONSTRAINT wmbs_fileset_unique UNIQUE (name) %s)""" % tablespaceIndex

        self.create["02wmbs_file_details"] = \
            """CREATE TABLE wmbs_file_details (
                 id          INTEGER NOT NULL,
                 lfn         VARCHAR(1250) NOT NULL,
                 filesize    INTEGER,
                 events      INTEGER,
                 first_event INTEGER      DEFAULT 0,
                 merged      CHAR(1) CHECK (merged IN ('0', '1' )) NOT NULL
                 ) %s""" % tablespaceTable

        self.indexes["01_pk_wmbs_file_details"] = \
            """ALTER TABLE wmbs_file_details ADD
                 (CONSTRAINT wmbs_file_details_pk PRIMARY KEY (id) %s)""" % tablespaceIndex

        self.indexes["02_pk_wmbs_file_details"] = \
            """ALTER TABLE wmbs_file_details ADD
                 (CONSTRAINT wmbs_fildetails_unique UNIQUE (lfn) %s)""" % tablespaceIndex

        self.create["03wmbs_fileset_files"] = \
            """CREATE TABLE wmbs_fileset_files (
                 fileid      INTEGER NOT NULL,
                 fileset     INTEGER NOT NULL,
                 insert_time INTEGER NOT NULL
                 ) %s""" % tablespaceTable

        # self.indexes["01_pk_wmbs_fileset_files"] = \
        #  """ALTER TABLE wmbs_fileset_files ADD
        #       (CONSTRAINT wmbs_fileset_files_pk PRIMARY KEY (fileid, fileset) %s)""" % tablespaceIndex

        self.constraints["01_fk_wmbs_fileset_files"] = \
            """ALTER TABLE wmbs_fileset_files ADD
                 (CONSTRAINT fk_filesetfiles_fileset FOREIGN KEY(fileset)
                    REFERENCES wmbs_fileset(id) ON DELETE CASCADE)"""

        self.constraints["01_idx_wmbs_fileset_files"] = \
            """CREATE INDEX wmbs_fileset_files_idx_fileset ON wmbs_fileset_files(fileset) %s""" % tablespaceIndex

        self.constraints["02_fk_wmbs_fileset_files"] = \
            """ALTER TABLE wmbs_fileset_files ADD
                 (CONSTRAINT fk_filesetfiles_file FOREIGN KEY(fileid)
                    REFERENCES wmbs_file_details(id) ON DELETE CASCADE)"""

        self.constraints["02_idx_wmbs_fileset_files"] = \
            """CREATE INDEX wmbs_fileset_files_idx_fileid ON wmbs_fileset_files(fileid) %s""" % tablespaceIndex

        self.create["04wmbs_file_parent"] = \
            """CREATE TABLE wmbs_file_parent (
                 child  INTEGER NOT NULL,
                 parent INTEGER NOT NULL
                 ) %s""" % tablespaceTable

        self.constraints["01_fk_wmbs_file_parent"] = \
            """ALTER TABLE wmbs_file_parent ADD
                 (CONSTRAINT fk_fileparent_parent FOREIGN KEY(parent)
                    REFERENCES wmbs_file_details(id) ON DELETE CASCADE)"""

        self.constraints["02_fk_wmbs_file_parent"] = \
            """ALTER TABLE wmbs_file_parent ADD
                 (CONSTRAINT fk_fileparent_child FOREIGN KEY(child)
                    REFERENCES wmbs_file_details(id) ON DELETE CASCADE)"""

        self.constraints["01_idx_wmbs_file_parent"] = \
            """CREATE INDEX wmbs_file_parent_parent ON wmbs_file_parent(parent) %s""" % tablespaceIndex

        self.constraints["02_idx_wmbs_file_parent"] = \
            """CREATE INDEX wmbs_file_parent_child ON wmbs_file_parent(child) %s""" % tablespaceIndex

        self.create["05wmbs_file_runlumi_map"] = \
            """CREATE TABLE wmbs_file_runlumi_map (
                 fileid INTEGER NOT NULL,
                 run    INTEGER NOT NULL,
                 lumi   INTEGER NOT NULL,
                 num_events INTEGER
                 ) %s""" % tablespaceTable

        self.indexes["01_pk_wmbs_file_runlumi_map"] = \
            """ALTER TABLE wmbs_file_runlumi_map ADD
                 (CONSTRAINT wmbs_file_runlumi_map_pk PRIMARY KEY (fileid, run, lumi) %s)""" % tablespaceIndex

        self.constraints["01_fk_wmbs_file_runlumi_map"] = \
            """ALTER TABLE wmbs_file_runlumi_map ADD
                 (CONSTRAINT fk_runlumi_file FOREIGN KEY (fileid)
                    REFERENCES wmbs_file_details(id) ON DELETE CASCADE)"""

        self.constraints["01_idx_wmbs_file_runlumi_map"] = \
            """CREATE INDEX wmbs_file_runlumi_map_fileid ON wmbs_file_runlumi_map(fileid) %s""" % tablespaceIndex

        self.create["05wmbs_location_state"] = \
            """CREATE TABLE wmbs_location_state (
               id   INTEGER NOT NULL,
               name VARCHAR(100) NOT NULL) %s""" % tablespaceTable

        self.indexes["01_pk_wmbs_location_state"] = \
            """ALTER TABLE wmbs_location_state ADD
                 (CONSTRAINT wmbs_location_state_pk PRIMARY KEY (id) %s)""" % tablespaceIndex

        self.create["06wmbs_location"] = \
            """CREATE TABLE wmbs_location (
                 id          INTEGER      NOT NULL,
                 site_name   VARCHAR(255) NOT NULL,
                 state       INTEGER      NOT NULL,
                 cms_name    VARCHAR(255),
                 ce_name     VARCHAR(255),
                 running_slots   INTEGER,
                 pending_slots   INTEGER,
                 plugin      VARCHAR(255),
                 state_time INTEGER DEFAULT 0
                 ) %s""" % tablespaceTable

        self.indexes["01_pk_wmbs_location"] = \
            """ALTER TABLE wmbs_location ADD
                 (CONSTRAINT wmbs_location_pk PRIMARY KEY (id) %s)""" % tablespaceIndex

        self.indexes["02_pk_wmbs_location"] = \
            """ALTER TABLE wmbs_location ADD
                 (CONSTRAINT wmbs_location_unique UNIQUE (site_name) %s)""" % tablespaceIndex

        self.constraints["01_fk_wmbs_location"] = \
            """ALTER TABLE wmbs_location ADD
                 (CONSTRAINT fk_location_state FOREIGN KEY (state)
                    REFERENCES wmbs_location_state(id))"""

        self.create["06wmbs_pnns"] = \
            """CREATE TABLE wmbs_pnns (
                 id   INTEGER,
                 pnn  VARCHAR(255))"""

        self.indexes["01_pk_wmbs_pnns"] = \
            """ALTER TABLE wmbs_pnns ADD
                 (CONSTRAINT wmbs_pnns_pk PRIMARY KEY (id) %s)""" % tablespaceIndex

        self.constraints["01_uq_wmbs_pnns"] = \
            """ALTER TABLE wmbs_pnns ADD
                 (CONSTRAINT wmbs_pnns_uq UNIQUE (pnn) %s)""" % tablespaceIndex

        self.create["07wmbs_location_pnns"] = \
            """CREATE TABLE wmbs_location_pnns (
                 location  INTEGER,
                 pnn       INTEGER
                 ) %s""" % tablespaceTable

        self.constraints["01_uq_wmbs_location_pnns"] = \
            """ALTER TABLE wmbs_location_pnns ADD
                 (CONSTRAINT wmbs_location_pnns_uq UNIQUE (location, pnn) %s)""" % tablespaceIndex

        self.constraints["02_fk_wmbs_location_pnns"] = \
            """ALTER TABLE wmbs_location_pnns ADD
                 (CONSTRAINT wmbs_location_pnns_location_fk FOREIGN KEY (location)
                   REFERENCES wmbs_location(id) ON DELETE CASCADE)"""

        self.constraints["03_fk_wmbs_location_pnns"] = \
            """ALTER TABLE wmbs_location_pnns ADD
                 (CONSTRAINT wmbs_location_pnns_pnn_fk FOREIGN KEY (pnn)
                   REFERENCES wmbs_pnns(id) ON DELETE CASCADE)"""

        self.create["07wmbs_users"] = \
            """CREATE TABLE wmbs_users (
               id         INTEGER      NOT NULL,
               cert_dn    VARCHAR(255) NOT NULL,
               name_hn    VARCHAR(255),
               owner      VARCHAR(255),
               grp        VARCHAR(255),
               group_name VARCHAR(255),
               role_name  VARCHAR(255)
               ) %s""" % tablespaceTable

        self.indexes["01_pk_wmbs_users"] = \
            """ALTER TABLE wmbs_users ADD
                 (CONSTRAINT wmbs_users_pk PRIMARY KEY (id) %s)""" % tablespaceIndex

        self.indexes["02_pk_wmbs_users"] = \
            """ALTER TABLE wmbs_users ADD
                (CONSTRAINT wmbs_users_unique UNIQUE (cert_dn, group_name, role_name) %s)""" % tablespaceIndex

        self.create["07wmbs_file_location"] = \
            """CREATE TABLE wmbs_file_location (
                 fileid   INTEGER NOT NULL,
                 pnn INTEGER NOT NULL
                 ) %s""" % tablespaceTable

        self.indexes["01_pk_wmbs_file_location"] = \
            """ALTER TABLE wmbs_file_location ADD
                 (CONSTRAINT wmbs_file_location_pk PRIMARY KEY (fileid, pnn) %s)""" % tablespaceIndex

        self.constraints["01_fk_wmbs_file_location"] = \
            """ALTER TABLE wmbs_file_location ADD
                (CONSTRAINT fk_location_file FOREIGN KEY(fileid)
                   REFERENCES wmbs_file_details(id) ON DELETE CASCADE)"""

        self.constraints["01_idx_wmbs_file_location"] = \
            """CREATE INDEX wmbs_file_location_fileid ON wmbs_file_location(fileid) %s""" % tablespaceIndex

        self.constraints["02_fk_wmbs_file_location"] = \
            """ALTER TABLE wmbs_file_location ADD
                (CONSTRAINT fk_location_location FOREIGN KEY(pnn)
                   REFERENCES wmbs_pnns(id) ON DELETE CASCADE)"""

        self.constraints["02_idx_wmbs_file_location"] = \
            """CREATE INDEX wmbs_file_location_location ON wmbs_file_location(pnn) %s""" % tablespaceIndex

        self.create["07wmbs_workflow"] = \
            """CREATE TABLE wmbs_workflow (
                 id    INTEGER      NOT NULL,
                 spec  VARCHAR(700) NOT NULL,
                 name  VARCHAR(255) NOT NULL,
                 task  VARCHAR(1250) NOT NULL,
                 type  VARCHAR(255),
                 owner INTEGER      NOT NULL,
                 alt_fs_close INTEGER NOT NULL,
                 injected INTEGER   DEFAULT 0,
                 priority INTEGER   DEFAULT 0
                 ) %s""" % tablespaceTable

        self.indexes["01_pk_wmbs_workflow"] = \
            """ALTER TABLE wmbs_workflow ADD
                 (CONSTRAINT wmbs_workflow_pk PRIMARY KEY (id) %s)""" % tablespaceIndex

        self.indexes["uniquewfname"] = \
            """ALTER TABLE wmbs_workflow ADD
                 (CONSTRAINT uniq_wf_name UNIQUE (name, task) %s)""" % tablespaceIndex

        self.indexes["02_fk_wmbs_workflow"] = \
            """ALTER TABLE wmbs_workflow ADD
                (CONSTRAINT fk_workflow_users FOREIGN KEY(owner)
                   REFERENCES wmbs_users(id) ON DELETE CASCADE)"""

        self.create["08wmbs_workflow_output"] = \
            """CREATE TABLE wmbs_workflow_output (
                 workflow_id           INTEGER      NOT NULL,
                 output_identifier     VARCHAR(255) NOT NULL,
                 output_fileset        INTEGER      NOT NULL,
                 merged_output_fileset INTEGER
                 ) %s""" % tablespaceTable

        self.constraints["01_fk_wmbs_workflow_output"] = \
            """ALTER TABLE wmbs_workflow_output ADD
                (CONSTRAINT fk_wfoutput_workflow FOREIGN KEY(workflow_id)
                   REFERENCES wmbs_workflow(id) ON DELETE CASCADE)"""

        self.constraints["01_idx_wmbs_workflow_output"] = \
            """CREATE INDEX idx_wmbs_workf_out_workflow ON wmbs_workflow_output(workflow_id) %s""" % tablespaceIndex

        self.constraints["02_fk_wmbs_workflow_output"] = \
            """ALTER TABLE wmbs_workflow_output ADD
                (CONSTRAINT fk_wfoutput_fileset FOREIGN KEY(output_fileset)
                   REFERENCES wmbs_fileset(id) ON DELETE CASCADE)"""

        self.constraints["03_fk_wmbs_workflow_output"] = \
            """ALTER TABLE wmbs_workflow_output ADD
                (CONSTRAINT fk_wfoutput_mfileset FOREIGN KEY(merged_output_fileset)
                   REFERENCES wmbs_fileset(id) ON DELETE CASCADE)"""

        self.constraints["02_idx_wmbs_workflow_output"] = \
            """CREATE INDEX idx_wmbs_workf_out_fileset ON wmbs_workflow_output(output_fileset) %s""" % tablespaceIndex

        self.constraints["03_idx_wmbs_workflow_output"] = \
            """CREATE INDEX idx_wmbs_workf_out_mfileset ON wmbs_workflow_output(merged_output_fileset) %s""" % tablespaceIndex

        self.create["08wmbs_sub_types"] = \
            """CREATE TABLE wmbs_sub_types (
                 id   INTEGER      NOT NULL,
                 name VARCHAR(255) NOT NULL,
                 priority INTEGER DEFAULT 0
                 ) %s""" % tablespaceTable

        self.indexes["01_pk_wmbs_sub_types"] = \
            """ALTER TABLE wmbs_sub_types ADD
                 (CONSTRAINT wmbs_sub_types_pk PRIMARY KEY (id) %s)""" % tablespaceIndex

        self.indexes["02_pk_wmbs_sub_types"] = \
            """ALTER TABLE wmbs_sub_types ADD
                 (CONSTRAINT wmbs_sub_types_uk UNIQUE (name) %s)""" % tablespaceIndex

        self.create["09wmbs_subscription"] = \
            """CREATE TABLE wmbs_subscription (
                 id          INTEGER      NOT NULL,
                 fileset     INTEGER      NOT NULL,
                 workflow    INTEGER      NOT NULL,
                 split_algo  VARCHAR(255) NOT NULL,
                 subtype     INTEGER      NOT NULL,
                 last_update INTEGER      NOT NULL,
                 finished    INTEGER      DEFAULT 0
                 ) %s""" % tablespaceTable

        self.indexes["01_pk_wmbs_subscription"] = \
            """ALTER TABLE wmbs_subscription ADD
                 (CONSTRAINT wmbs_subscription_pk PRIMARY KEY (id) %s)""" % tablespaceIndex

        self.constraints["01_fk_wmbs_subscription"] = \
            """ALTER TABLE wmbs_subscription ADD
                 (CONSTRAINT fk_subs_fileset FOREIGN KEY(fileset)
                    REFERENCES wmbs_fileset(id) ON DELETE CASCADE)"""

        self.constraints["01_idx_wmbs_subscription"] = \
            """CREATE INDEX idx_wmbs_subscription_fileset ON wmbs_subscription(fileset) %s""" % tablespaceIndex

        self.constraints["02_fk_wmbs_subscription"] = \
            """ALTER TABLE wmbs_subscription ADD
                 (CONSTRAINT fk_sub_types FOREIGN KEY(subtype)
                    REFERENCES wmbs_sub_types(id) ON DELETE CASCADE)"""

        self.constraints["02_idx_wmbs_subscription"] = \
            """CREATE INDEX idx_wmbs_subscription_subtype ON wmbs_subscription(subtype) %s""" % tablespaceIndex

        self.constraints["03_fk_wmbs_subscription"] = \
            """ALTER TABLE wmbs_subscription ADD
                 (CONSTRAINT fk_subs_workflow FOREIGN KEY(workflow)
                    REFERENCES wmbs_workflow(id) ON DELETE CASCADE)"""

        self.constraints["03_idx_wmbs_subscription"] = \
            """CREATE INDEX idx_wmbs_subscription_workflow ON wmbs_subscription(workflow) %s""" % tablespaceIndex

        self.create["10wmbs_subscription_validation"] = \
            """CREATE TABLE wmbs_subscription_validation (
               subscription_id INTEGER NOT NULL,
               location_id     INTEGER NOT NULL,
               valid           INTEGER)"""

        self.indexes["01_pk_wmbs_sub_val"] = \
            """ALTER TABLE wmbs_subscription_validation ADD
                 (CONSTRAINT wmbs_sub_val_pk PRIMARY KEY (subscription_id, location_id) %s)""" % tablespaceIndex

        self.constraints["01_fk_wmbs_sub_val"] = \
            """ALTER TABLE wmbs_subscription_validation ADD
                (CONSTRAINT fk_sub_val FOREIGN KEY(subscription_id)
                   REFERENCES wmbs_subscription(id) ON DELETE CASCADE)"""

        self.constraints["02_fk_wmbs_sub_val"] = \
            """ALTER TABLE wmbs_subscription_validation ADD
                (CONSTRAINT fk2_sub_val FOREIGN KEY(location_id)
                   REFERENCES wmbs_location(id) ON DELETE CASCADE)"""

        self.create["10wmbs_sub_files_acquired"] = \
            """CREATE TABLE wmbs_sub_files_acquired (
                 subscription INTEGER NOT NULL,
                 fileid       INTEGER NOT NULL
                 ) %s""" % tablespaceTable

        self.indexes["01_pk_wmbs_sub_files_acquired"] = \
            """ALTER TABLE wmbs_sub_files_acquired ADD
                 (CONSTRAINT wmbs_sub_files_acquired_pk PRIMARY KEY (subscription, fileid) %s)""" % tablespaceIndex

        self.constraints["01_fk_wmbs_sub_files_acquired"] = \
            """ALTER TABLE wmbs_sub_files_acquired ADD
                 (CONSTRAINT fk_subsacquired_sub FOREIGN KEY (subscription)
                    REFERENCES wmbs_subscription(id) ON DELETE CASCADE)"""

        self.constraints["02_fk_wmbs_sub_files_acquired"] = \
            """ALTER TABLE wmbs_sub_files_acquired ADD
                 (CONSTRAINT fk_subsacquired_file FOREIGN KEY (fileid)
                    REFERENCES wmbs_file_details(id) ON DELETE CASCADE)"""

        self.constraints["01_idx_wmbs_sub_files_acquired"] = \
            """CREATE INDEX idx_wmbs_sub_files_acq_sub ON wmbs_sub_files_acquired(subscription) %s""" % tablespaceIndex

        self.constraints["02_idx_wmbs_sub_files_acquired"] = \
            """CREATE INDEX idx_wmbs_sub_files_acq_file ON wmbs_sub_files_acquired(fileid) %s""" % tablespaceIndex

        self.create["10wmbs_sub_files_available"] = \
            """CREATE TABLE wmbs_sub_files_available (
                 subscription INTEGER NOT NULL,
                 fileid       INTEGER NOT NULL
                 ) %s""" % tablespaceTable

        self.indexes["01_pk_wmbs_sub_files_available"] = \
            """ALTER TABLE wmbs_sub_files_available ADD
                 (CONSTRAINT wmbs_sub_files_available_pk PRIMARY KEY (subscription, fileid) %s)""" % tablespaceIndex

        self.constraints["01_fk_wmbs_sub_files_available"] = \
            """ALTER TABLE wmbs_sub_files_available ADD
                 (CONSTRAINT fk_subsavailable_sub FOREIGN KEY (subscription)
                    REFERENCES wmbs_subscription(id) ON DELETE CASCADE)"""

        self.constraints["02_fk_wmbs_sub_files_available"] = \
            """ALTER TABLE wmbs_sub_files_available ADD
                 (CONSTRAINT fk_subsavailable_file FOREIGN KEY (fileid)
                    REFERENCES wmbs_file_details(id) ON DELETE CASCADE)"""

        self.constraints["01_idx_wmbs_sub_files_available"] = \
            """CREATE INDEX idx_wmbs_sub_files_ava_sub ON wmbs_sub_files_available(subscription) %s""" % tablespaceIndex

        self.constraints["02_idx_wmbs_sub_files_available"] = \
            """CREATE INDEX idx_wmbs_sub_files_ava_file ON wmbs_sub_files_available(fileid) %s""" % tablespaceIndex

        self.create["11wmbs_sub_files_failed"] = \
            """CREATE TABLE wmbs_sub_files_failed (
                 subscription INTEGER NOT NULL,
                 fileid       INTEGER NOT NULL
                 ) %s""" % tablespaceTable

        self.indexes["01_pk_wmbs_sub_files_failed"] = \
            """ALTER TABLE wmbs_sub_files_failed ADD
                 (CONSTRAINT wmbs_sub_files_failed_pk PRIMARY KEY (subscription, fileid) %s)""" % tablespaceIndex

        self.constraints["01_fk_wmbs_sub_files_failed"] = \
            """ALTER TABLE wmbs_sub_files_failed ADD
                 (CONSTRAINT fk_subsfailed_sub FOREIGN KEY (subscription)
                    REFERENCES wmbs_subscription(id) ON DELETE CASCADE)"""

        self.constraints["02_fk_wmbs_sub_files_failed"] = \
            """ALTER TABLE wmbs_sub_files_failed ADD
                 (CONSTRAINT fk_subsfailed_file FOREIGN KEY (fileid)
                    REFERENCES wmbs_file_details(id) ON DELETE CASCADE)"""

        self.constraints["01_idx_wmbs_sub_files_failed"] = \
            """CREATE INDEX idx_wmbs_sub_files_fail_sub ON wmbs_sub_files_failed(subscription) %s""" % tablespaceIndex

        self.constraints["02_idx_wmbs_sub_files_failed"] = \
            """CREATE INDEX idx_wmbs_sub_files_fail_file ON wmbs_sub_files_failed(fileid) %s""" % tablespaceIndex

        self.create["12wmbs_sub_files_complete"] = \
            """CREATE TABLE wmbs_sub_files_complete (
                 subscription INTEGER NOT NULL,
                 fileid       INTEGER NOT NULL
                 ) %s""" % tablespaceTable

        self.indexes["01_pk_wmbs_sub_files_complete"] = \
            """ALTER TABLE wmbs_sub_files_complete ADD
                 (CONSTRAINT wmbs_sub_files_complete_pk PRIMARY KEY (subscription, fileid) %s)""" % tablespaceIndex

        self.constraints["01_fk_wmbs_sub_files_complete"] = \
            """ALTER TABLE wmbs_sub_files_complete ADD
                 (CONSTRAINT fk_subscomplete_sub FOREIGN KEY (subscription)
                    REFERENCES wmbs_subscription(id) ON DELETE CASCADE)"""

        self.constraints["02_fk_wmbs_sub_files_complete"] = \
            """ALTER TABLE wmbs_sub_files_complete ADD
                 (CONSTRAINT fk_subscomplete_file FOREIGN KEY (fileid)
                    REFERENCES wmbs_file_details(id) ON DELETE CASCADE)"""

        self.constraints["01_idx_wmbs_sub_files_complete"] = \
            """CREATE INDEX idx_wmbs_sub_files_comp_sub ON wmbs_sub_files_complete(subscription) %s""" % tablespaceIndex

        self.constraints["02_idx_wmbs_sub_files_complete"] = \
            """CREATE INDEX idx_wmbs_sub_files_comp_file ON wmbs_sub_files_complete(fileid) %s""" % tablespaceIndex

        self.create["13wmbs_jobgroup"] = \
            """CREATE TABLE wmbs_jobgroup (
                 id           INTEGER       NOT NULL,
                 subscription INTEGER       NOT NULL,
                 guid         VARCHAR(255),
                 output       INTEGER,
                 last_update  INTEGER       NOT NULL,
                 location     INTEGER
                 ) %s""" % tablespaceTable

        self.indexes["01_pk_wmbs_jobgroup"] = \
            """ALTER TABLE wmbs_jobgroup ADD
                 (CONSTRAINT wmbs_jobgroup_pk PRIMARY KEY (id) %s)""" % tablespaceIndex

        self.indexes["03_pk_wmbs_jobgroup"] = \
            """ALTER TABLE wmbs_jobgroup ADD
                 (CONSTRAINT wmbs_jobgroup_unique2 UNIQUE (guid) %s)""" % tablespaceIndex

        self.constraints["01_fk_wmbs_jobgroup"] = \
            """ALTER TABLE wmbs_jobgroup ADD
                 (CONSTRAINT fk_jobgroup_subscription FOREIGN KEY (subscription)
                    REFERENCES wmbs_subscription(id) ON DELETE CASCADE)"""

        self.constraints["02_fk_wmbs_jobgroup"] = \
            """ALTER TABLE wmbs_jobgroup ADD
                 (CONSTRAINT fk_jobgroup_fileset FOREIGN KEY (output)
                    REFERENCES wmbs_fileset(id) ON DELETE CASCADE)"""

        self.constraints["01_idx_wmbs_sub_jobgroup"] = \
            """CREATE INDEX idx_wmbs_jobgroup_sub ON wmbs_jobgroup(subscription) %s""" % tablespaceIndex

        self.create["14wmbs_job_state"] = \
            """CREATE TABLE wmbs_job_state (
                 id   INTEGER      NOT NULL,
                 name VARCHAR(100) NOT NULL
                 ) %s""" % tablespaceTable

        self.indexes["01_pk_wmbs_job_state"] = \
            """ALTER TABLE wmbs_job_state ADD
                 (CONSTRAINT wmbs_job_state_pk PRIMARY KEY (id) %s)""" % tablespaceIndex

        self.create["15wmbs_job"] = \
            """CREATE TABLE wmbs_job (
                 id           INTEGER       NOT NULL,
                 jobgroup     INTEGER       NOT NULL,
                 name         VARCHAR(255),
                 state        INTEGER       NOT NULL,
                 state_time   INTEGER       NOT NULL,
                 retry_count  INTEGER       DEFAULT 0,
                 couch_record VARCHAR(255),
                 location     INTEGER,
                 outcome      INTEGER       DEFAULT 0,
                 cache_dir    VARCHAR(1250)  DEFAULT 'None',
                 fwjr_path    VARCHAR(1250)
                 ) %s""" % tablespaceTable

        self.indexes["01_pk_wmbs_job"] = \
            """ALTER TABLE wmbs_job ADD
                 (CONSTRAINT wmbs_job_pk PRIMARY KEY (id) %s)""" % tablespaceIndex

        self.indexes["02_pk_wmbs_job"] = \
            """ALTER TABLE wmbs_job ADD
                 (CONSTRAINT wmbs_job_uk UNIQUE (name) %s)""" % tablespaceIndex

        self.constraints["01_fk_wmbs_job"] = \
            """ALTER TABLE wmbs_job ADD
                 (CONSTRAINT wmbs_job_fk_jobgroup FOREIGN KEY (jobgroup)
                    REFERENCES wmbs_jobgroup(id) ON DELETE CASCADE)"""

        self.constraints["01_idx_wmbs_job"] = \
            """CREATE INDEX idx_wmbs_job_jobgroup ON wmbs_job(jobgroup) %s""" % tablespaceIndex

        self.constraints["02_fk_wmbs_job"] = \
            """ALTER TABLE wmbs_job ADD
                 (CONSTRAINT fk_location FOREIGN KEY (location)
                    REFERENCES wmbs_location(id))"""

        self.constraints["02_idx_wmbs_job"] = \
            """CREATE INDEX idx_wmbs_job_loc ON wmbs_job(location) %s""" % tablespaceIndex

        self.constraints["03_fk_wmbs_job"] = \
            """ALTER TABLE wmbs_job ADD
                 (CONSTRAINT fk_state FOREIGN KEY (state)
                    REFERENCES wmbs_job_state(id))"""

        self.constraints["03_idx_wmbs_job"] = \
            """CREATE INDEX idx_wmbs_job_state ON wmbs_job(state) %s""" % tablespaceIndex

        self.create["16wmbs_job_assoc"] = \
            """CREATE TABLE wmbs_job_assoc (
                 job    INTEGER NOT NULL,
                 fileid INTEGER NOT NULL
                 ) %s""" % tablespaceTable

        self.indexes["01_pk_wmbs_job_assoc"] = \
            """ALTER TABLE wmbs_job_assoc ADD
                 (CONSTRAINT wmbs_job_assoc_pk PRIMARY KEY (job, fileid) %s)""" % tablespaceIndex

        self.constraints["01_fk_wmbs_job_assoc"] = \
            """ALTER TABLE wmbs_job_assoc ADD
                 (CONSTRAINT fk_jobassoc_job FOREIGN KEY (job)
                    REFERENCES wmbs_job(id) ON DELETE CASCADE)"""

        self.constraints["01_idx_wmbs_job_assoc"] = \
            """CREATE INDEX idx_wmbs_job_assoc_job ON wmbs_job_assoc(job) %s""" % tablespaceIndex

        self.constraints["02_fk_wmbs_job_assoc"] = \
            """ALTER TABLE wmbs_job_assoc ADD
                 (CONSTRAINT fk_jobassoc_file FOREIGN KEY (fileid)
                    REFERENCES wmbs_file_details(id) ON DELETE CASCADE)"""

        self.constraints["02_idx_wmbs_job_assoc"] = \
            """CREATE INDEX idx_wmbs_job_assoc_file ON wmbs_job_assoc(fileid) %s""" % tablespaceIndex

        self.create["17wmbs_job_mask"] = \
            """CREATE TABLE wmbs_job_mask (
                 job           INTEGER  NOT NULL,
                 FirstEvent    INTEGER,
                 LastEvent     INTEGER,
                 FirstLumi     INTEGER,
                 LastLumi      INTEGER,
                 FirstRun      INTEGER,
                 LastRun       INTEGER,
                 inclusivemask CHAR(1) NOT NULL
                 ) %s""" % tablespaceTable

        self.constraints["01_fk_wmbs_job_mask"] = \
            """ALTER TABLE wmbs_job_mask ADD
                 (CONSTRAINT fk_mask_job FOREIGN KEY (job)
                    REFERENCES wmbs_job(id) ON DELETE CASCADE)"""

        self.constraints["01_idx_wmbs_job_mask"] = \
            """CREATE INDEX idx_wmbs_job_mask_job ON wmbs_job_mask(job) %s""" % tablespaceIndex

        self.create["18wmbs_checksum_type"] = \
            """CREATE TABLE wmbs_checksum_type (
                id            INTEGER,
                type          VARCHAR(255)
                ) %s""" % tablespaceTable

        self.indexes["01_pk_wmbs_checksum_type"] = \
            """ALTER TABLE wmbs_checksum_type ADD
                 (CONSTRAINT wmbs_checksum_type_pk PRIMARY KEY (id) %s)""" % tablespaceIndex

        self.create["19wmbs_file_checksums"] = \
            """CREATE TABLE wmbs_file_checksums (
                fileid        INTEGER,
                typeid        INTEGER,
                cksum         VARCHAR(100)
                ) %s""" % tablespaceTable

        self.indexes["02_uk_wmbs_file_checksums"] = \
            """ALTER TABLE wmbs_file_checksums ADD
                 (CONSTRAINT wmbs_file_checksums_uk UNIQUE (fileid, typeid) %s)""" % tablespaceIndex

        self.constraints["02_fk_wmbs_file_checksums"] = \
            """ALTER TABLE wmbs_file_checksums ADD
                 (CONSTRAINT fk_filechecksums_cktype FOREIGN KEY (typeid)
                    REFERENCES wmbs_checksum_type(id) ON DELETE CASCADE)"""

        self.constraints["01_idx_wmbs_file_checksums"] = \
            """CREATE INDEX idx_wmbs_file_checksums_type ON wmbs_file_checksums(typeid) %s""" % tablespaceIndex

        self.constraints["03_fk_wmbs_file_checksums"] = \
            """ALTER TABLE wmbs_file_checksums ADD
                 (CONSTRAINT fk_filechecksums_file FOREIGN KEY (fileid)
                    REFERENCES wmbs_file_details(id) ON DELETE CASCADE)"""

        self.constraints["01_idx_wmbs_file_checksums"] = \
            """CREATE INDEX idx_wmbs_file_checksums_file ON wmbs_file_checksums(fileid) %s""" % tablespaceIndex

        # Workunit table for tracking individual lumis, indices come from CreateWMBSBase.py

        self.create["21wmbs_workunit"] = (
            'CREATE TABLE wmbs_workunit('
            ' id INTEGER NOT NULL,'
            ' taskid INTEGER NOT NULL,'
            ' retry_count INTEGER DEFAULT 0,'
            ' last_unit_count INTEGER NOT NULL,'
            ' last_submit_time INTEGER NOT NULL,'
            ' status INTEGER DEFAULT 0,'
            ' PRIMARY KEY(id)'
            ') %s' % tablespaceTable
        )
        self.constraints['01_fk_wmbs_workunit'] = ('ALTER TABLE wmbs_workunit ADD '
                                                   '(CONSTRAINT wmbs_workunit_fk_taskid'
                                                   ' FOREIGN KEY(taskid) REFERENCES wmbs_workflow(id)'
                                                   ' ON DELETE CASCADE)')

        # Association table between jobs and workunits, indices come from CreateWMBSBase.py
        self.create["22wmbs_job_workunit_assoc"] = (
            'CREATE TABLE wmbs_job_workunit_assoc ('
            ' job    INTEGER NOT NULL,'
            ' workunit INTEGER NOT NULL'
            ') %s' % tablespaceTable

        )
        self.constraints['01_fk_wmbs_job_workunit_assoc'] = ('ALTER TABLE wmbs_job_workunit_assoc ADD '
                                                             '(CONSTRAINT wmbs_job_wu_assoc_fk_job'
                                                             ' FOREIGN KEY(job) REFERENCES wmbs_job(id)'
                                                             ' ON DELETE CASCADE)')
        self.constraints['02_fk_wmbs_job_workunit_assoc'] = ('ALTER TABLE wmbs_job_workunit_assoc ADD '
                                                             '(CONSTRAINT wmbs_job_wu_assoc_fk_wu'
                                                             ' FOREIGN KEY(workunit) REFERENCES wmbs_workunit(id)'
                                                             ' ON DELETE CASCADE)')

        # Association table between workunits and file/run/lumi, indices come from CreateWMBSBase.py
        self.create["23wmbs_frl_workunit_assoc"] = (
            'CREATE TABLE wmbs_frl_workunit_assoc ('
            ' workunit INTEGER NOT NULL,'
            ' firstevent INTEGER DEFAULT 0,'
            ' lastevent INTEGER DEFAULT 0,'
            ' fileid  INTEGER NOT NULL,'
            ' run     INTEGER NOT NULL,'
            ' lumi    INTEGER NOT NULL,'
            ' PRIMARY KEY(workunit, fileid, run, lumi)'
            ') %s' % tablespaceTable
        )
        self.constraints['01_fk_wmbs_frl_workunit_assoc'] = ('ALTER TABLE wmbs_frl_workunit_assoc ADD '
                                                             '(CONSTRAINT wmbs_frl_wu_assoc_fk_wu'
                                                             ' FOREIGN KEY(workunit) REFERENCES wmbs_workunit(id)'
                                                             ' ON DELETE CASCADE)')
        self.constraints['02_fk_wmbs_frl_workunit_assoc'] = ('ALTER TABLE wmbs_frl_workunit_assoc ADD '
                                                             '(CONSTRAINT wmbs_frl_wu_assoc_fk_frl'
                                                             ' FOREIGN KEY(fileid, run, lumi) '
                                                             ' REFERENCES wmbs_file_runlumi_map(fileid, run, lumi)'
                                                             ' ON DELETE CASCADE)')
        for jobState in Transitions().states():
            jobStateQuery = """INSERT INTO wmbs_job_state(id, name) VALUES
                               (wmbs_job_state_SEQ.nextval, '%s')""" % jobState
            self.inserts["job_state_%s" % jobState] = jobStateQuery

        self.subTypes = [("Processing", 0), ("Merge", 4), ("Harvesting", 5), ("Cleanup", 1),
                         ("LogCollect", 2), ("Skim", 3), ("Production", 0)]
        for pair in self.subTypes:
            subTypeQuery = """INSERT INTO wmbs_sub_types (id, name, priority)
                              VALUES (wmbs_sub_types_SEQ.nextval, '%s', %d)""" % (pair[0], pair[1])
            self.inserts["wmbs_sub_types_%s" % pair[0]] = subTypeQuery

        locationStates = ["Normal", "Down", "Draining", "Aborted"]

        for i in locationStates:
            locationStateQuery = """INSERT INTO wmbs_location_state (id, name)
                                    VALUES (wmbs_location_state_SEQ.nextval, '%s')""" % i
            self.inserts["wmbs_location_state_%s" % i] = locationStateQuery

        checksumTypes = ["cksum", "adler32", "md5"]
        for i in checksumTypes:
            checksumTypeQuery = \
                """INSERT INTO wmbs_checksum_type (id, type)
                   VALUES (wmbs_checksum_type_SEQ.nextval, '%s')""" % i
            self.inserts["wmbs_checksum_type_%s" % i] = checksumTypeQuery

        j = 50
        for i in self.sequence_tables:
            seqname = '%s_SEQ' % i
            self.create["%s%s" % (j, seqname)] = \
                """CREATE SEQUENCE %s start with 1 increment by 1 nomaxvalue cache 100""" % seqname



#!/usr/bin/env python
"""
_FNALImpl_

Implementation of StageOutImpl interface for FNAL

"""
from builtins import range

from future import standard_library
standard_library.install_aliases()

import logging
import os

from WMCore.Storage.Plugins.LCGImpl import LCGImpl
from WMCore.Storage.StageOutImplV2 import StageOutImplV2
from subprocess import getoutput

_CheckExitCodeOption = True
checkPathsCount = 5
checkPaths = ['/lustre/unmerged/', '/lustre/temp/', '/store/unmerged/', '/store/temp/', '/store/user/meloam/lustre']
checkPathsReplace = ['root://cmseos.fnal.gov//lustre/unmerged/', 'root://cmseos.fnal.gov//lustre/temp/',
                     'root://cmseos.fnal.gov//lustre/unmerged/', 'root://cmseos.fnal.gov//lustre/temp/',
                     'root://cmseos.fnal.gov//eos/uscms/store/user/meloam/wmagent-lustre/']
srmPaths = ['/store/temp/user/', '/store/user/']
dccpPaths = ['/store/user/meloam/t4/']

# For testing
envScript = "/opt/d-cache/dcap/bin/setenv-cmsprod.sh"


# envScript = "/uscms_data/d3/meloam/wmagentstuff/site-packages/setenv-meloam.sh"

def pnfsPfn2(pfn):
    """
    _pnfsPfn2_

    Convert a dcap PFN to a PNFS PFN

    """
    # only change PFN on remote storage
    if pfn.find('/pnfs/') == -1:
        return pfn

    pfnSplit = pfn.split("WAX/11/store/", 1)[1]
    filePath = "/pnfs/cms/WAX/11/store/%s" % pfnSplit

    # Find vetoed paths first
    for path in srmPaths:
        if pfn.find(path) != -1:
            return filePath

    # handle lustre location
    for i in range(checkPathsCount):
        if pfn.find(checkPaths[i]) != -1:
            pfnSplit = pfn.split(checkPaths[i], 1)[1]
            filePath = "%s%s" % (checkPathsReplace[i], pfnSplit)
    return filePath


class FNALImpl(StageOutImplV2):
    """
    _FNALImpl_

    Implement interface for dcache door based dccp command

    """

    def __init__(self):

        StageOutImplV2.__init__(self)

        # Create and hold onto a srm implementation in case we need it
        self.srmImpl = LCGImpl()

    def storageMethod(self, PFN):
        """
        Figure out which paths use DCAP, lustre, or SRM for access
        """

        method = 'dccp'
        for path in checkPaths:
            if PFN.find(path) != -1:
                method = 'lustre'

        # Over ride a few paths for srm
        for path in srmPaths:
            if PFN.find(path) != -1:
                method = 'srm'

        # overrides for testing
        for path in dccpPaths:
            if PFN.find(path) != -1:
                method = 'dccp'

        if PFN.find('/store/user/meloam/lustre/') != -1:
            method = 'lustre'

        logging.debug("Using method %s for PFN %s", method, PFN)
        return method

    def substituteLustrePath(self, pfn):
        for i in range(checkPathsCount):
            if pfn.find(checkPaths[i]) != -1:
                pfnSplit = pfn.split(checkPaths[i], 1)[1]
                pfn = "%s%s" % (checkPathsReplace[i], pfnSplit)
        return pfn

    def dcapToPNFS(self, pfn):
        if pfn.find('/pnfs/') == -1:
            raise RuntimeError("dcapToPNFS called on a non-dcap/pnfs URL: %s" % pfn)

        pfnSplit = pfn.split("WAX/11/store/", 1)[1]
        filePath = "/pnfs/cms/WAX/11/store/%s" % pfnSplit
        return filePath

    def createOutputDirectoryDCAP(self, targetPFN):
        """
        _createOutputDirectory_

        Create a dir for the target pfn by translating it to
        a /pnfs name and calling mkdir

        PFN will be of the form:
        dcap://cmsdca.fnal.gov:22125/pnfs/fnal.gov/usr/cms/WAX/11/store/blah

        We need to convert that into /pnfs/cms/WAX/11/store/blah, as it
        will be seen from the worker node

        Unmerged PFN will be of the form:
        /lustre/unmerged

        we don't need to convert it, just mkdir.


        """
        logging.info("createOutputDirectoryDCAP called on %s", targetPFN)
        # only create dir on remote storage
        if targetPFN.find('/pnfs/') == -1:
            return

        filePath = self.dcapToPNFS(targetPFN)
        directory = os.path.dirname(filePath)
        logging.info("Creating path: %s", directory)
        if not os.path.exists(directory):
            os.makedirs(directory)

    def createOutputDirectoryEOS(self, targetPFN):
        logging.info("createOutputDirectoryEOS called on %s", targetPFN)
        # only create dir on remote storage
        if targetPFN.find('root://cmseos.fnal.gov') == -1:
            return

        filePath = targetPFN.split("root://cmseos.fnal.gov/", 1)[1]
        directory = os.path.dirname(filePath)
        logging.info("Creating path: %s", directory)
        if not os.path.exists(directory):
            os.makedirs(directory)

    def createSourceName(self, protocol, pfn):
        """
        createTargetName

        generate the target PFN

        """
        # if not pfn.startswith("srm"):
        #    return pfn

        method = self.storageMethod(pfn)

        if method == 'srm':
            return self.srmImpl.createSourceName(protocol, pfn)
        elif method == 'dccp':
            logging.info("Translating PFN for dcache: %s", pfn)
            pfn = pfn.split("/store/")[1]
            dcacheDoor = getoutput(". " + envScript + "\n" + "/opt/d-cache/dcap/bin/select_RdCapDoor.sh")
            pfn = "%s%s" % (dcacheDoor, pfn)
            logging.info("  Translated PFN: %s", pfn)
        elif method == 'lustre':
            logging.info("Translating PFN for lustre: %s", pfn)
            pfn = self.substituteLustrePath(pfn)
            logging.info("  Translated PFN: %s", pfn)
        else:
            raise RuntimeError("Unknown method found in createSourceName: %s" % method)

        return pfn

    def doDelete(self, pfnToRemove, pnn, command, options, protocol):
        """
        _removeFile_

        CleanUp pfn provided

        """

        method = self.storageMethod(pfnToRemove)

        if method == 'srm':
            return self.srmImpl.doDelete(pfnToRemove, pnn, command, options, protocol)
        elif method == 'dccp':
            pfnSplit = pfnToRemove.split("/11/store/", 1)[1]
            filePath = "/pnfs/cms/WAX/11/store/%s" % pfnSplit
            if os.path.exists(filePath):
                logging.info("Removing file: %s", filePath)
                os.unlink(filePath)
        elif method == 'lustre':
            pfnToRemove = self.substituteLustrePath(pfnToRemove)
            filePath = pfnToRemove.split("root://cmseos.fnal.gov/", 1)[1]
            if os.path.exists(filePath):
                logging.info("Removing file: %s", filePath)
                os.unlink(filePath)
        else:
            raise RuntimeError("Unsupported storage method in doDelete: %s" % method)

    def doTransfer(self, fromPfn, toPfn, stageOut, pnn, command, options, protocol, checksum):
        """
            performs a transfer. stageOut tells you which way to go. returns the new pfn or
            raises on failure. StageOutError (and inherited exceptions) are for expected errors
            such as temporary connection failures. Anything else will be handled as an unexpected
            error and skip retrying with this plugin

            if stageOut is true:
                The fromPfn is the LOCAL FILE NAME on the node, without file://
                the toPfn is the target PFN, mapped from the LFN using the TFC or overrrides
            if stageOut is false:
                The toPfn is the LOCAL FILE NAME on the node, without file://
                the fromPfn is the source PFN, mapped from the LFN using the TFC or overrrides

            this behavior is because most transfer commands will switch their direction
            simply by swapping the order of the arguments. the stageOut flag is provided
            however, because sometimes you want to pass different command line args
        """
        if stageOut:
            localPFN = fromPfn
            remotePFN = toPfn
        else:
            localPFN = toPfn
            remotePFN = fromPfn

        method = self.storageMethod(remotePFN)
        sourceMethod = self.storageMethod(localPFN)

        if method == 'srm' or sourceMethod == 'srm':
            return self.srmImpl.doTransfer(fromPfn, toPfn, stageOut, \
                                           pnn, command, options, protocol, checksum)

        # transfer with lustre
        elif method == 'lustre':
            sourcePFN = self.substituteLustrePath(fromPfn)
            targetPFN = self.substituteLustrePath(toPfn)

            # there is no xrd module yet ...
            if stageOut:
                self.createOutputDirectoryEOS(targetPFN)
            else:
                self.createOutputDirectoryEOS(sourcePFN)

            optionsStr = ""
            if options != None:
                optionsStr = str(options)

            original_size = os.stat(sourcePFN)[6]
            useChecksum = (checksum != None and 'adler32' in checksum and not self.stageIn)
            if useChecksum:
                checksum['adler32'] = "%08x" % int(checksum['adler32'], 16)

                # non-functional in 3.3.1 xrootd clients due to bug
                # result += "-ODeos.targetsize=$LOCAL_SIZE\&eos.checksum=%s " % checksums['adler32']

                # therefor embed information into target URL
                targetPFN += "\?eos.targetsize=%s\&eos.checksum=%s" % (original_size, checksum['adler32'])

            result = "#!/bin/sh\n"
            result += "xrdcp -d 0 %s %s %s\n" % (optionsStr, sourcePFN, targetPFN)
            try:
                self.runCommandFailOnNonZero(result)
            except:
                logging.info("xrdcp failed, removing failed file")
                if not stageOut and os.path.exists(pnfsPfn2(targetPFN)):
                    logging.info("unlinking %s", pnfsPfn2(targetPFN))
                    os.unlink(pnfsPfn2(targetPFN))
                else:
                    self.doDelete(targetPFN, pnn, command, options, protocol)
                raise


        elif method == 'dccp':
            if stageOut:
                self.createOutputDirectoryDCAP(targetPFN)
                targetPFN = self.createSourceName(protocol, targetPFN)
            else:
                self.createOutputDirectoryDCAP(sourcePFN)
                sourcePFN = self.createSourceName(protocol, sourcePFN)

            optionsStr = ""
            if options != None:
                optionsStr = str(options)
            result = "#!/bin/bash\n"
            result += ". " + envScript + "\n"
            result += "dccp -o 86400  -d 0 -X -role=cmsprod %s %s %s" % (optionsStr, sourcePFN, targetPFN)
            try:
                self.runCommandFailOnNonZero(result)
            except:
                logging.info("dccp failed, removing failed file")
                if not stageOut and os.path.exists(pnfsPfn2(targetPFN)):
                    logging.info("unlinking %s", pnfsPfn2(targetPFN))
                    os.unlink(pnfsPfn2(targetPFN))
                else:
                    self.doDelete(targetPFN, pnn, command, options, protocol)
                raise

            # //
            # //  CRC check
            # //
            try:
                if stageOut:
                    self.runCommandFailOnNonZero("/opt/d-cache/dcap/bin/check_dCachefilecksum.sh %s %s" \
                                                 % (pnfsPfn2(targetPFN), sourcePFN))
                else:
                    self.runCommandFailOnNonZero("/opt/d-cache/dcap/bin/check_dCachefilecksum.sh %s %s" \
                                                 % (pnfsPfn2(sourcePFN), targetPFN))

            except:
                logging.info("DCCP CRC Check failed, removing failed file")
                self.doDelete(targetPFN, pnn, command, options, protocol)
                raise
            return



#!/usr/bin/env python
"""
_IsCompleteOnRun_

Oracle implementation of Subscriptions.IsCompleteOnRun
"""

from WMCore.WMBS.MySQL.Subscriptions.IsCompleteOnRun import IsCompleteOnRun as IsCompleteOnRunMySQL

class IsCompleteOnRun(IsCompleteOnRunMySQL):
    pass



#!/usr/bin/env python
"""
_FailedJobsByTask_

Monitoring DAO classes for Jobs in WMBS
"""
__all__ = []



from WMCore.WMBS.MySQL.Monitoring.FailedJobsByTask \
  import FailedJobsByTask as FailedJobsByTaskMySQL

class FailedJobsByTask(FailedJobsByTaskMySQL):
    pass



#!/usr/bin/env python
"""
_AlgoDatasetAssoc_

MySQL implementation of DBSBuffer.AlgoDatasetAssoc
"""




from WMCore.Database.DBFormatter import DBFormatter

class AlgoDatasetAssoc(DBFormatter):
    """
    _AlgoDatasetAssoc_

    Associtate an algorithm to a dataset and return the ID of the assocation.
    """
    sql = """INSERT INTO dbsbuffer_algo_dataset_assoc (algo_id, dataset_id)
               SELECT (SELECT id FROM dbsbuffer_algo WHERE app_name = :app_name AND
                         app_ver = :app_ver AND app_fam = :app_fam AND
                         pset_hash = :pset_hash) AS algo_id,
                      (SELECT id FROM dbsbuffer_dataset WHERE path = :path) AS dataset_id
               FROM DUAL WHERE NOT EXISTS
                 (SELECT * FROM dbsbuffer_algo_dataset_assoc WHERE algo_id =
                   (SELECT id FROM dbsbuffer_algo WHERE app_name = :app_name AND
                      app_ver = :app_ver AND app_fam = :app_fam AND
                      pset_hash = :pset_hash) AND dataset_id =
                   (SELECT id FROM dbsbuffer_dataset WHERE path = :path))"""

    idsql = """SELECT id FROM dbsbuffer_algo_dataset_assoc WHERE algo_id =
                 (SELECT id FROM dbsbuffer_algo WHERE app_name = :app_name AND
                    app_ver = :app_ver AND app_fam = :app_fam AND
                    pset_hash = :pset_hash) AND dataset_id =
                 (SELECT id FROM dbsbuffer_dataset WHERE path = :path)"""

    def execute(self, appName = None, appVer = None, appFam = None,
                psetHash = None, datasetPath = None, conn = None,
                transaction = False):
        binds = {"app_name": appName, "app_ver": appVer, "app_fam": appFam,
                 "pset_hash": psetHash, "path": datasetPath}

        self.dbi.processData(self.sql, binds, conn = conn,
                             transaction = transaction)

        result = self.dbi.processData(self.idsql, binds, conn = conn,
                                      transaction = transaction)
        formattedResult = self.formatDict(result)
        if len(formattedResult) == 1:
            return formattedResult[0]["id"]

        return None



#!/usr/bin/env python
"""
_LoadBlocks_

MySQL implementation of LoadBlocks
"""

from WMCore.Database.DBFormatter import DBFormatter

class LoadBlocks(DBFormatter):

    sql = """SELECT dbsbuffer_block.blockname as blockname,
                    dbsbuffer_block.create_time as create_time,
                    dbsbuffer_block.status AS status,
                    dbsbuffer_dataset.path AS datasetpath,
                    dbsbuffer_location.pnn AS location
              FROM dbsbuffer_block
              INNER JOIN dbsbuffer_dataset ON
                dbsbuffer_dataset.id = dbsbuffer_block.dataset_id
              INNER JOIN dbsbuffer_location ON
                dbsbuffer_location.id = dbsbuffer_block.location
              WHERE dbsbuffer_block.blockname = :blockname
              """

    def format(self, result):

        tmpList = self.formatDict(result)
        blockList = []
        for tmp in tmpList:
            blockList.append( { 'block_name' : tmp['blockname'],
                                'creation_date' : tmp['create_time'],
                                'origin_site_name' : tmp['location'],
                                'datasetpath' : tmp['datasetpath'],
                                'status' : tmp['status'] } )

        return blockList

    def execute(self, blocknames, conn = None, transaction = False):
        """
        Take a list of blocknames and use them to load
        the blocks.
        """
        binds = []
        for name in blocknames:
            binds.append({'blockname': name})
        result = self.dbi.processData(self.sql, binds, conn = conn,
                                      transaction = transaction)
        return self.format(result)



#!/usr/bin/env python
"""
_ListOpen_

Oracle implementation of Fileset.ListOpen
"""

__all__ = []



from WMCore.WMBS.MySQL.Fileset.ListOpen import ListOpen as ListOpenFilesetMySQL

class ListOpen(ListOpenFilesetMySQL):
    sql = ListOpenFilesetMySQL.sql
